# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import numpy as np
from scipy.sparse import csr_matrix as sparse, eye, vstack, hstack
from pandapower.pypower.idx_bus import BASE_KV
from pandapower.pypower.idx_brch import F_BUS, T_BUS, BR_STATUS

SIGMA = 5.670374419e-8
# ALPHA = 4.03e-3
ALPHA = 4e-3


def calc_r_theta_from_t_rise(net, t_rise_degree_celsius):
    """
    Calculate thermal resistance of the conductors from an assumed or calculated temperature rise.
    The calculation is implemented according to Frank et al.

    Parameters
    ----------
    net : pandapowerNet
    t_rise_degree_celsius : array
        temperature rise of the conductor

    Returns
    -------
    r_theta_kelvin_per_mw : array
        Thermal resistance of the conductors R_{\\Theta}

    References
    ----------
    S. Frank, J. Sexauer and S. Mohagheghi, "Temperature-Dependent Power Flow," in IEEE Transactions on Power Systems,
    vol. 28, no. 4, pp. 4007-4018, Nov. 2013, doi: 10.1109/TPWRS.2013.2266409.
    """
    r_for_t_rated_rise = net.line.r_ohm_per_km * (1 + net.line.alpha * t_rise_degree_celsius) * \
                         net.line.length_km / net.line.parallel
    p_rated_loss_mw = np.square(net.line.max_i_ka * np.sqrt(3)) * r_for_t_rated_rise
    r_theta_kelvin_per_mw = t_rise_degree_celsius / p_rated_loss_mw
    return r_theta_kelvin_per_mw


def calc_i_square_p_loss(branch, tdpf_lines, g, b, Vm, Va):
    """
    Calculate squared current and the active power losses.

    Parameters
    ----------
    branch: np.array(complex)
        ppc["branch"]
    tdpf_lines : np.array(bool)
        array that defines which lines are relevant for TDPF
    g : array
    b : array
    Vm : array
        Bus voltage magnitude (p.u.)
    Va : array
        Bus voltage angle (rad)

    Returns
    -------
    i_square_pu : array
        squared current
    p_loss_pu : array
        active power losses
    """
    i = branch[tdpf_lines, F_BUS].real.astype(np.int64)
    j = branch[tdpf_lines, T_BUS].real.astype(np.int64)

    A = np.square(Vm[i]) + np.square(Vm[j]) - 2 * Vm[i] * Vm[j] * np.cos(Va[i] - Va[j])

    # i_square_pu matches net.res_line.i_ka only if c_nf_per_km == 0
    i_square_pu = (np.square(g) + np.square(b)) * A

    # p_loss_pu here is correct, matches net.res_line.pl_mw
    p_loss_pu = g * A

    return i_square_pu, p_loss_pu


def calc_r_theta(t_air_pu, a0, a1, a2, i_square_pu, p_loss_pu):
    """
    Calculate thermal resistance using the thermal model from Ngoko et al.

    Parameters
    ----------
    t_air_pu : array
        Air temperature in p.u.
    a0 : array
        constant term of the thermal model
    a1 : array
        linear term of the thermal model
    a2 : array
        quadratic term of the thermal model
    i_square_pu : array
        squared current in p.u.
    p_loss_pu : array
        active power losses in p.u.

    Returns
    -------
    r_theta_pu : array
        Thermal resistance of the conductors R_{\\Theta}

    References
    ----------
    S. Frank, J. Sexauer and S. Mohagheghi, "Temperature-Dependent Power Flow," in IEEE Transactions on Power Systems,
    vol. 28, no. 4, pp. 4007-4018, Nov. 2013, doi: 10.1109/TPWRS.2013.2266409.

    B. Ngoko, H. Sugihara and T. Funaki, "A Temperature Dependent Power Flow Model Considering Overhead Transmission
    Line Conductor Thermal Inertia Characteristics," 2019 IEEE International Conference on Environment and
    Electrical Engineering and 2019 IEEE Industrial and Commercial Power Systems Europe (EEEIC / I&CPS Europe),
    2019, pp. 1-6, doi: 10.1109/EEEIC.2019.8783234.
    """
    t_rise_pu = a0 + a1 * i_square_pu + a2 * np.square(i_square_pu) - t_air_pu
    r_theta_pu = t_rise_pu / np.where(p_loss_pu == 0, 1e-6, p_loss_pu)
    return r_theta_pu


def calc_T_frank(p_loss_pu, t_air_pu, r_theta_pu, tdpf_delay_s, T0, tau):
    """
    Calculate overhead line temperature according to the method from Frank et al.
    The calculation of the overhead line temperature is based on their thermal resistance.

    Parameters
    ----------
    p_loss_pu : array
        active power losses in p.u.
    t_air_pu : array
        Air temperature in p.u.
    r_theta_pu : array
        Thermal resistance of the conductors R_{\\Theta}
    tdpf_delay_s : float, None
        Delay for the consideration of thermal inertia in seconds. Describes the time passed after a change
        of current in overhead lines that causes a change of temperature.
        Example: tdpf_delay_s = 0 means there is no change in temperature;
        tdpf_delay_s = np.inf leads to obtaining steady-state temperature (also default behavior if tdpf_delay_s = None)
    T0 : array, None
        initial temperature of overhead lines
    tau : array, None
        time constant of the overhead lines; describes the time after a current change after which the temperature
        reaches approx. 63.2 % of the steady-state value

    Returns
    -------
    t_transient : array
        Temperature of the overhead lines, either steady-state or corresponding to the time delay tdpf_delay_s

    References
    ----------
    S. Frank, J. Sexauer and S. Mohagheghi, "Temperature-Dependent Power Flow," in IEEE Transactions on Power Systems,
    vol. 28, no. 4, pp. 4007-4018, Nov. 2013, doi: 10.1109/TPWRS.2013.2266409.
    """
    t_ss = t_air_pu + r_theta_pu * p_loss_pu

    if tdpf_delay_s is None:
        return t_ss

    t_transient = t_ss - (t_ss - T0) * np.exp(-tdpf_delay_s / tau)
    return t_transient


def calc_T_ngoko(i_square_pu, a0, a1, a2, tdpf_delay_s, T0, tau):
    """
    Calculate the overhead line temperature with the approach from Ngoko et al.
    The calculation of the overhead line temperature is based on the simplified model that
    includes a constant term, a linear coefficient and a quadratic coefficient.

    Parameters
    ----------
    i_square_pu : array
        squared current in p.u.
    a0 : array
        constant term of the thermal model
    a1 : array
        linear term of the thermal model
    a2 : array
        quadratic term of the thermal model
    tdpf_delay_s : float, None
        Delay for the consideration of thermal inertia in seconds. Describes the time passed after a change
        of current in overhead lines that causes a change of temperature.
        Example: tdpf_delay_s = 0 means there is no change in temperature;
        tdpf_delay_s = np.inf leads to obtaining steady-state temperature (also default behavior if tdpf_delay_s = None)
    T0 : array, None
        initial temperature of overhead lines
    tau : array, None
        time constant of the overhead lines; describes the time after a current change after which the temperature
        reaches approx. 63.2 % of the steady-state value

    Returns
    -------
    t_transient : array
        Temperature of the overhead lines, either steady-state or corresponding to the time delay tdpf_delay_s

    References
    ----------
    B. Ngoko, H. Sugihara and T. Funaki, "A Temperature Dependent Power Flow Model Considering Overhead Transmission
    Line Conductor Thermal Inertia Characteristics," 2019 IEEE International Conference on Environment and
    Electrical Engineering and 2019 IEEE Industrial and Commercial Power Systems Europe (EEEIC / I&CPS Europe),
    2019, pp. 1-6, doi: 10.1109/EEEIC.2019.8783234.
    """
    t_ss = a0 + a1 * i_square_pu + a2 * np.square(i_square_pu)

    if tdpf_delay_s is None:
        return t_ss

    t_transient = t_ss - (t_ss - T0) * np.exp(-tdpf_delay_s / tau)
    return t_transient


def calc_a0_a1_a2_tau(t_air_pu, t_max_pu, t_ref_pu, r_ref_ohm_per_m, conductor_outer_diameter_m,
                      mc_joule_per_m_k, wind_speed_m_per_s, wind_angle_degree, s_w_per_square_meter,
                      alpha_pu=ALPHA, solar_absorptivity=0.5, emissivity=0.5, T_base=1, i_base_a=1):
    """
    Calculate the coefficients for the simplified thermal model according to Ngoko et al.

    Parameters
    ----------
    t_air_pu : array
        Air temperature in p.u.
    t_max_pu : array
        max. rated temperature of the overhead lines
    t_ref_pu : array
        rated temperature at which the reference (datasheet) resistance is provided
    r_ref_ohm_per_m : array
        reference (datasheet) specific resistance of the overhead lines
    conductor_outer_diameter_m : array
        outer diameter of the overhead line conductors (diameter of 1 individual conductor of the overhead line)
    mc_joule_per_m_k : array
        specific thermal capacitance of the overhead line:
        mass per unit length m [kg/m] multiplied by the specific thermal capacity c [J/kg • K]
    wind_speed_m_per_s : array
        wind speed in m/s
    wind_angle_degree : array
        wind angle of attack
    s_w_per_square_meter : array
        solar radiation in W/m²
    alpha_pu : array
        temperature coefficient of resistance in p.u. - alpha multiplied by T_base
    solar_absorptivity : array
    emissivity : array
    T_base : array
        base value for T for calculating T in p.u.
    i_base_a : array
        base value for current for calculating I in p.u.

    Returns
    -------
    a0 : array
        constant term of the thermal model
    a1 : array
        linear term of the thermal model
    a2 : array
        quadratic term of the thermal model
    tau : array
        time constant of the overhead lines; describes the time after a current change after which the temperature
        reaches approx. 63.2 % of the steady-state value

    References
    ----------
    B. Ngoko, H. Sugihara and T. Funaki, "A Temperature Dependent Power Flow Model Considering Overhead Transmission
    Line Conductor Thermal Inertia Characteristics," 2019 IEEE International Conference on Environment and
    Electrical Engineering and 2019 IEEE Industrial and Commercial Power Systems Europe (EEEIC / I&CPS Europe),
    2019, pp. 1-6, doi: 10.1109/EEEIC.2019.8783234.
    """
    # alpha here is expected to be for T in pu (alpha multiplied by T_base)
    r_amb_ohm_per_m = r_ref_ohm_per_m * (1 + alpha_pu * (t_air_pu - t_ref_pu))
    r_max_ohm_per_m = r_ref_ohm_per_m * (1 + alpha_pu * (t_max_pu - t_ref_pu))

    h_r = 4 * np.pi * conductor_outer_diameter_m * SIGMA * emissivity * (t_air_pu * T_base + 273) ** 3
    kappa = 6 * np.pi * conductor_outer_diameter_m * SIGMA * emissivity * (t_air_pu * T_base + 273) ** 2

    h_c = calc_h_c(conductor_outer_diameter_m, wind_speed_m_per_s, wind_angle_degree, t_air_pu * T_base)

    k2 = r_max_ohm_per_m / (h_r + h_c + kappa)

    a0 = t_air_pu + (solar_absorptivity * conductor_outer_diameter_m * s_w_per_square_meter) / (h_r + h_c) / T_base
    a1 = r_amb_ohm_per_m / (h_r + h_c) / T_base * np.square(i_base_a)
    a2 = k2 / (h_r + h_c) * (alpha_pu / T_base * r_ref_ohm_per_m - kappa * k2) / T_base * np.power(i_base_a, 4)
    # a2 = a1 / (h_r + h_c) * (alpha/T_base * r_ref_ohm_per_m - kappa * a1) / T_base * np.power(i_base_a, 4)

    # rho = 2710  # kg/m³ # density of aluminum
    # c = 1.309e6  # J/kg°C
    tau = mc_joule_per_m_k / (h_r + h_c)

    return a0, a1, a2, tau


def calc_h_c(conductor_outer_diameter_m, v_m_per_s, wind_angle_degree, t_air_degree_celsius):
    rho_air = 101325 / (287.058 * (t_air_degree_celsius + 273))  # pressure 1 atm. / (R_specific * T)
    rho_air_relative = 1.  # relative air density
    # r_f = 0.05 # roughness of conductors
    r_f = 0.1 # roughness of conductors
    w = rho_air * conductor_outer_diameter_m * v_m_per_s

    K = np.where(v_m_per_s < 0.5, 0.55,
                 np.where(v_m_per_s < 24,
                          0.42 + 0.68 * np.sin(np.deg2rad(wind_angle_degree)) ** 1.08,
                          0.42 + 0.58 * np.sin(np.deg2rad(wind_angle_degree)) ** 0.9))

    h_cfl = 8.74 * K * w ** 0.471

    h_cfh = 13.44 * K * w ** 0.633 if r_f <= 0.05 else 20.89 * K * w ** 0.8

    h_cn = 8.1 * conductor_outer_diameter_m ** 0.75

    h_c = np.maximum(np.maximum(h_cfl, h_cfh), h_cn)

    return h_c

#
# def calc_a0_a1_a2_old(t_air_degree_celsius, t_max, r_ref_ohm_per_m, conductor_outer_diameter_m, v_m_per_s, wind_angle_degree, s_w_per_square_meter=300, alpha=ALPHA, solar_absorptivity=0.5, emissivity=0.5):
#     r_amb_ohm_per_m = calc_r_temp(r_ref_ohm_per_m, t_air_degree_celsius)
#     r_max_ohm_per_m = calc_r_temp(r_ref_ohm_per_m, t_max)
#     h_r = 4 * SIGMA * emissivity * (t_air_degree_celsius + 273) ** 3
#     # h_r = 4 * np.pi * conductor_outer_diameter_m * SIGMA * emissivity * (t_air_degree_celsius + 273) ** 3
#     h_c = calc_h_c(conductor_outer_diameter_m, v_m_per_s, wind_angle_degree, t_air_degree_celsius)
#     # R0 = 1 / (np.pi * conductor_outer_diameter_m * (h_r + h_c))
#     R0 = 1 / (h_r + h_c)
#     kappa = 6 * np.pi * conductor_outer_diameter_m * SIGMA * emissivity * (t_air_degree_celsius + 273) ** 2
#     R1 = 1 / (np.pi * conductor_outer_diameter_m * (h_r + h_c + kappa * (t_max - t_air_degree_celsius)))
#     Ps = solar_absorptivity * conductor_outer_diameter_m * s_w_per_square_meter
#     #a0 = t_air_degree_celsius + R0 * Ps
#     a0 = t_air_degree_celsius + (solar_absorptivity * conductor_outer_diameter_m * s_w_per_square_meter) / (h_r + h_c)
#     a1 = R0 * r_amb_ohm_per_m
#     # a2 = R0 * R1 * r_max_ohm_per_m * (alpha * r_ref_ohm_per_m - kappa)
#     a2 = a1 / (h_r + h_c) * (alpha * r_ref_ohm_per_m - kappa * a1)
#     return a0, a1, a2
#
#
# def calc_h_c_old(conductor_outer_diameter_m, v_m_per_s, wind_angle_degree):
#     B1 = 0.178
#     n1 = 0.633
#     # rho_air = 1.2041  # kg_per_m3
#     rho_air = 1.  # relative air density
#
#     if v_m_per_s < 0.5:
#         K = 0.55
#     elif wind_angle_degree < 24:
#         K = 0.42 + 0.68 * np.sin(np.deg2rad(wind_angle_degree)) ** 1.08
#     else:
#         K = 0.42 + 0.58 * np.sin(np.deg2rad(wind_angle_degree)) ** 0.9
#
#     # K = 1.194 - np.cos(np.deg2rad(wind_angle_degree)) + (0.194 * np.cos(2 * np.deg2rad(wind_angle_degree))) + (0.368 * np.sin(2 * np.deg2rad(wind_angle_degree)))
#
#     h_cl = 3.07 / conductor_outer_diameter_m * K * (rho_air * conductor_outer_diameter_m * v_m_per_s) ** 0.471
#
#     h_ch = B1 / conductor_outer_diameter_m * K * (rho_air * conductor_outer_diameter_m * v_m_per_s) ** n1
#
#     h_c = np.maximum(h_cl, h_ch)
#
#     return h_c
#
#
# def calc_tau_old(R0, q_mm2, rho, c, h_r, h_c):
#     # rho = 2710  # kg/m³ # density of aluminum
#     # c = 1.309e6  # J/kg°C
#     q_m2 = q_mm2 * 1e-6
#     m_kg_per_m = rho_kg_per_m3 * area_m2
#     tau = m_kg_per_m * c / (h_r + h_c)
#     return tau


def create_J_tdpf(branch, tdpf_lines, alpha_pu, r_ref_pu, pvpq, pq, pvpq_lookup, pq_lookup, tau, tdpf_delay_s, Vm, Va,
                  r_theta_pu, J, r, x, g):
    """
             | J11 = dP/dd     J12 = dP/dV     J13 = dP/dT  |
             | (N-1)x(N-1)     (N-1)x(M)       (N-1)x(R)    |
             |                                              |
             | J21 = dQ/dd     J22 = dQ/dV     J23 = dQ/dT  |
             | (M)x(N-1)       (M)x(M)         (M)x(R)      |
             |                                              |
             | J31 = ddT/dd    J32 = ddT/dV    J33 = ddT/dT |
             | (R)x(N-1)       (R)x(M)         (R)x(R)      |

            N = Number of buses
            M = Number of PQ buses
            R = Number temperature-dependent branches

    References
    ----------
    S. Frank, J. Sexauer and S. Mohagheghi, "Temperature-Dependent Power Flow," in IEEE Transactions on Power Systems,
    vol. 28, no. 4, pp. 4007-4018, Nov. 2013, doi: 10.1109/TPWRS.2013.2266409.

    B. Ngoko, H. Sugihara and T. Funaki, "A Temperature Dependent Power Flow Model Considering Overhead Transmission
    Line Conductor Thermal Inertia Characteristics," 2019 IEEE International Conference on Environment and
    Electrical Engineering and 2019 IEEE Industrial and Commercial Power Systems Europe (EEEIC / I&CPS Europe),
    2019, pp. 1-6, doi: 10.1109/EEEIC.2019.8783234.
    """
    C = np.ones_like(tdpf_lines, dtype=np.float64)
    if tdpf_delay_s is not None and tdpf_delay_s != np.inf:
        C *= (1 - np.exp(-tdpf_delay_s / tau))

    dg_dT = (np.square(x) - np.square(r)) * alpha_pu * r_ref_pu / np.square(np.square(r) + np.square(x))
    db_dT = 2 * x * g * alpha_pu * r_ref_pu / (np.square(r) + np.square(x))

    in_pq_f = np.isin(branch[tdpf_lines, F_BUS].real.astype(np.int64), pq)
    in_pq_t = np.isin(branch[tdpf_lines, T_BUS].real.astype(np.int64), pq)
    in_pvpq_f = np.isin(branch[tdpf_lines, F_BUS].real.astype(np.int64), pvpq)
    in_pvpq_t = np.isin(branch[tdpf_lines, T_BUS].real.astype(np.int64), pvpq)

    # todo: optimize and speed-up the code for the matrices (write numba versions)
    J13 = create_J13(branch, tdpf_lines, in_pvpq_f, in_pvpq_t, pvpq, pvpq_lookup, Vm, Va, dg_dT, db_dT)
    J23 = create_J23(branch, tdpf_lines, in_pq_f, in_pq_t, pq, pq_lookup, Vm, Va, dg_dT, db_dT)
    J31 = create_J31(branch, tdpf_lines, in_pvpq_f, in_pvpq_t, pvpq, pvpq_lookup, Vm, Va, C, r_theta_pu, g)
    J32 = create_J32(branch, tdpf_lines, in_pq_f, in_pq_t, pq, pq_lookup, Vm, Va, C, r_theta_pu, g)
    J33 = create_J33(branch, tdpf_lines, r_theta_pu, Vm, Va, dg_dT)

    Jright = vstack([sparse(J13), sparse(J23)], format="csr")
    Jbtm = hstack([sparse(J31), sparse(J32), sparse(J33)], format="csr")
    JJ = vstack([hstack([J, Jright]), Jbtm], format="csr")
    return JJ


def calc_I(Sf, bus, f_bus, V):
    If = 1e3 * abs(Sf) / (abs(V[f_bus]) * bus[f_bus, BASE_KV].astype(np.float64)) / np.sqrt(3)
    return If


def calc_g_b(r, x):
    g = r / (np.square(r) + np.square(x))
    b = -x / (np.square(r) + np.square(x))
    return g, b


def get_S_flows(branch, Yf, Yt, baseMVA, V):
    br = branch[:, BR_STATUS].real.astype(bool)
    f_bus = np.real(branch[br, F_BUS]).astype(np.int64)
    Sf = V[f_bus] * np.conj(Yf[br, :] * V) * baseMVA
    # complex power injected at "to" bus
    t_bus = np.real(branch[br, T_BUS]).astype(np.int64)
    St = V[t_bus] * np.conj(Yt[br, :] * V) * baseMVA
    return Sf, St, f_bus, t_bus

#
# def calc_AB(branch, tdpf_lines, pvpq, pvpq_lookup, Va, Vm):
#     # A = np.zeros(shape=(len(pvpq), len(pvpq)))
#     # B = np.zeros(shape=(len(pvpq), len(pvpq)))
#     A = np.zeros(shape=(len(Vm), len(Vm)))
#     B = np.zeros(shape=(len(Vm), len(Vm)))
#
#     # figure out the indexing for the pv buses:
#     for br in tdpf_lines:
#         f, t = branch[br, [F_BUS, T_BUS]].real.astype(np.int64)
#         #m = pvpq_lookup[f]
#         #i = pvpq_lookup[t]
#         m = f
#         i = t
#         A[m, i] = np.square(Vm[m]) - Vm[m] * Vm[i] * np.cos(Va[m] - Va[i])
#         A[i, m] = np.square(Vm[i]) - Vm[m] * Vm[i] * np.cos(Va[i] - Va[m])
#         B[m, i] = Vm[m] * Vm[i] * np.sin(Va[m] - Va[i])
#         B[i, m] = Vm[m] * Vm[i] * np.sin(Va[i] - Va[m])
#
#     # for bus in pvpq:
#     #     m = int(pvpq_lookup[bus])
#     #     if bus in branch[:, F_BUS].real.astype(np.int64):
#     #         other = branch[branch[:, F_BUS] == bus, T_BUS].real.astype(np.int64)
#     #         i = pvpq_lookup[other]
#     #     elif bus in branch[:, T_BUS].real.astype(np.int64):
#     #         other = branch[branch[:, T_BUS] == bus, F_BUS].real.astype(np.int64)
#     #         i = pvpq_lookup[other]
#     #     else:
#     #         continue
#     #
#     #     A[m, i] = np.square(Vm[m]) - Vm[m] * Vm[i] * np.cos(Va[m]-Va[i])
#     #     B[m, i] = Vm[m] * Vm[i] * np.sin(Va[m]-Va[i])
#
#     return A, B


def create_J13(branch, tdpf_lines, in_pvpq_f, in_pvpq_t, pvpq, pvpq_lookup, Vm, Va, dg_dT, db_dT):
    """
         | J11 = dP/dd     J12 = dP/dV     J13 = dP/dT  |
         | (N-1)x(N-1)     (N-1)x(M)       (N-1)x(R)    |
         |                                              |
         | J21 = dQ/dd     J22 = dQ/dV     J23 = dQ/dT  |
         | (M)x(N-1)       (M)x(M)         (M)x(R)      |
         |                                              |
         | J31 = ddT/dd    J32 = ddT/dV    J33 = ddT/dT |
         | (R)x(N-1)       (R)x(M)         (R)x(R)      |

    N = Number of buses
    M = Number of PQ buses
    R = Number temperature-dependent branches

    shape = (len(branch), len(bus))
    :param b:
    :param g:
    :param dg_dT:
    :param db_dT:
    :param Vm:
    :param Va:
    :param pvpq_lookup:

    """
    nrow = len(pvpq)
    ncol = len(branch)
    J13 = np.zeros(shape=(nrow, ncol), dtype=np.float64)

    # for m in pvpq:
    #     mm = pvpq_lookup[m]
    #     # m = bus
    #     #for ij in range(ncol):
    #     for ij_lookup, ij in enumerate(tdpf_lines):
    #         i = branch[ij, F_BUS].real.astype(np.int64)
    #         j = branch[ij, T_BUS].real.astype(np.int64)
    #
    #         if m == i:
    #             n = j
    #         elif m == j:
    #             n = i
    #         else:
    #             continue
    #
    #         A_mn = Vm[m] ** 2 - Vm[m] * Vm[n] * np.cos(Va[m] - Va[n])
    #         B_mn = Vm[m] * Vm[n] * np.sin(Va[m] - Va[n])
    #         # p_mn = g[ij_lookup] * A_mn - b[ij_lookup] * B_mn
    #         # J13[mm, ij] = alpha[ij] * r_ref[ij] * g[ij] * (A_mn / r[ij] - 2 * p_mn)
    #         J13[mm, ij] = A_mn * dg_dT[ij_lookup] - B_mn * db_dT[ij_lookup]

    mf = np.r_[branch[tdpf_lines[in_pvpq_f], F_BUS].real.astype(np.int64)]
    mt = np.r_[branch[tdpf_lines[in_pvpq_t], T_BUS].real.astype(np.int64)]
    nf = np.r_[branch[tdpf_lines[in_pvpq_f], T_BUS].real.astype(np.int64)]
    nt = np.r_[branch[tdpf_lines[in_pvpq_t], F_BUS].real.astype(np.int64)]

    for in_pq, m, n in ((in_pvpq_f, mf, nf), (in_pvpq_t, mt, nt)):
        pq_j = pvpq_lookup[m]
        A_mn = Vm[m] ** 2 - Vm[m] * Vm[n] * np.cos(Va[m] - Va[n])
        B_mn = Vm[m] * Vm[n] * np.sin(Va[m] - Va[n])
        J13[pq_j, tdpf_lines[in_pq]] = A_mn * dg_dT[in_pq] - B_mn * db_dT[in_pq]

    return J13


def create_J23(branch, tdpf_lines, in_pq_f, in_pq_t, pq, pq_lookup, Vm, Va, dg_dT, db_dT):
    """
         | J11 = dP/dd     J12 = dP/dV     J13 = dP/dT  |
         | (N-1)x(N-1)     (N-1)x(M)       (N-1)x(R)    |
         |                                              |
         | J21 = dQ/dd     J22 = dQ/dV     J23 = dQ/dT  |
         | (M)x(N-1)       (M)x(M)         (M)x(R)      |
         |                                              |
         | J31 = ddT/dd    J32 = ddT/dV    J33 = ddT/dT |
         | (R)x(N-1)       (R)x(M)         (R)x(R)      |

        N = Number of buses
        M = Number of PQ buses
        R = Number temperature-dependent branches

    shape = (len(bus), len(branch))
    :param g:
    :param b:
    :param dg_dT:
    :param db_dT:
    :param Vm:
    :param Va:
    :param pq_lookup:

    """

    ncol = len(branch)
    nrow = len(pq)
    J23 = np.zeros(shape=(nrow, ncol), dtype=np.float64)

    if nrow == 0:
        return J23

    # for m in pq:
    #     mm = pq_lookup[m]
    #     # m = bus
    #     #for ij in range(ncol):
    #     for ij_lookup, ij in enumerate(tdpf_lines):
    #         i = branch[ij, F_BUS].real.astype(np.int64)
    #         j = branch[ij, T_BUS].real.astype(np.int64)
    #
    #         if m == i:
    #             n = j
    #         elif m == j:
    #             n = i
    #         else:
    #             continue
    #
    #         A_mn = Vm[m] ** 2 - Vm[m] * Vm[n] * np.cos(Va[m] - Va[n])
    #         B_mn = Vm[m] * Vm[n] * np.sin(Va[m] - Va[n])
    #         q_mn = -b[ij_lookup] * A_mn - g[ij_lookup] * B_mn
    #         # J13[mm, ij] = alpha[ij] * r_ref[ij] * g[ij] * (B_mn / r[ij] - 2 * q_mn)
    #         J23[mm, ij] = - A_mn * db_dT[ij_lookup] - B_mn * dg_dT[ij_lookup]

    mf = np.r_[branch[tdpf_lines[in_pq_f], F_BUS].real.astype(np.int64)]
    mt = np.r_[branch[tdpf_lines[in_pq_t], T_BUS].real.astype(np.int64)]
    nf = np.r_[branch[tdpf_lines[in_pq_f], T_BUS].real.astype(np.int64)]
    nt = np.r_[branch[tdpf_lines[in_pq_t], F_BUS].real.astype(np.int64)]

    for in_pq, m, n in ((in_pq_f, mf, nf), (in_pq_t, mt, nt)):
        pq_j = pq_lookup[m]
        A_mn = Vm[m] ** 2 - Vm[m] * Vm[n] * np.cos(Va[m] - Va[n])
        B_mn = Vm[m] * Vm[n] * np.sin(Va[m] - Va[n])
        J23[pq_j, tdpf_lines[in_pq]] = - A_mn * db_dT[in_pq] - B_mn * dg_dT[in_pq]

    return J23


def create_J31(branch, tdpf_lines, in_pvpq_f, in_pvpq_t, pvpq, pvpq_lookup, Vm, Va, C, r_theta_pu, g):
    """
         | J11 = dP/dd     J12 = dP/dV     J13 = dP/dT  |
         | (N-1)x(N-1)     (N-1)x(M)       (N-1)x(R)    |
         |                                              |
         | J21 = dQ/dd     J22 = dQ/dV     J23 = dQ/dT  |
         | (M)x(N-1)       (M)x(M)         (M)x(R)      |
         |                                              |
         | J31 = ddT/dd    J32 = ddT/dV    J33 = ddT/dT |
         | (R)x(N-1)       (R)x(M)         (R)x(R)      |

    N = Number of buses
    M = Number of PQ buses
    R = Number temperature-dependent branches

    shape = (len(branch), len(m))
    branch elements by row, m elements by column
    :param g:
    :param b:
    :param r_theta_pu:
    :param Vm:
    :param Va:
    :param pvpq_lookup:

    """

    nrow = len(branch)
    ncol = len(pvpq)

    J31 = np.zeros(shape=(nrow, ncol), dtype=np.float64)

    # #for ij in range(nrow):
    # for ij_lookup, ij in enumerate(tdpf_lines):
    #     i = branch[ij, F_BUS].real.astype(np.int64)
    #     j = branch[ij, T_BUS].real.astype(np.int64)
    #     for m in pvpq:
    #         mm = pvpq_lookup[m]
    #         if m == i:
    #             n = j
    #             sign = 1
    #         elif m == j:
    #             n = i
    #             sign = -1
    #         else:
    #             continue
    #
    #         B_mn = Vm[m] * Vm[n] * np.sin(Va[m] - Va[n])
    #         # J31[ij, mm] = sign * (g[ij] ** 2 + b[ij] ** 2) * C[ij] * B_mn
    #         J31_old[ij, mm] = - 2 * r_theta_pu[ij_lookup] * g[ij_lookup] * B_mn * C[ij_lookup]

    mf = np.r_[branch[tdpf_lines[in_pvpq_f], F_BUS].real.astype(np.int64)]
    mt = np.r_[branch[tdpf_lines[in_pvpq_t], T_BUS].real.astype(np.int64)]
    nf = np.r_[branch[tdpf_lines[in_pvpq_f], T_BUS].real.astype(np.int64)]
    nt = np.r_[branch[tdpf_lines[in_pvpq_t], F_BUS].real.astype(np.int64)]

    for in_pvpq, m, n in ((in_pvpq_f, mf, nf), (in_pvpq_t, mt, nt)):
        pq_j = pvpq_lookup[m]
        J31[tdpf_lines[in_pvpq], pq_j] = - 2 * r_theta_pu[in_pvpq] * g[in_pvpq] * Vm[m] * Vm[n] * np.sin(Va[m] - Va[n]) * C[in_pvpq]

    return J31


def create_J32(branch, tdpf_lines, in_pq_f, in_pq_t, pq, pq_lookup, Vm, Va, C, r_theta_pu, g):
    """
         | J11 = dP/dd     J12 = dP/dV     J13 = dP/dT  |
         | (N-1)x(N-1)     (N-1)x(M)       (N-1)x(R)    |
         |                                              |
         | J21 = dQ/dd     J22 = dQ/dV     J23 = dQ/dT  |
         | (M)x(N-1)       (M)x(M)         (M)x(R)      |
         |                                              |
         | J31 = ddT/dd    J32 = ddT/dV    J33 = ddT/dT |
         | (R)x(N-1)       (R)x(M)         (R)x(R)      |

        N = Number of buses
        M = Number of PQ buses
        R = Number temperature-dependent branches

    shape = (len(branch), len(m))
    branch elements by ij, m elements by column
    :param g:
    :param b:
    :param r_theta_pu:
    :param pq_lookup:

    """

    nrow = len(branch)
    ncol = len(pq)

    J32 = np.zeros(shape=(nrow, ncol), dtype=np.float64)

    if ncol == 0:
        return J32

    # # #for ij in range(nrow):
    # for ij_lookup, ij in enumerate(tdpf_lines):
    #     i = branch[ij, F_BUS].real.astype(np.int64)
    #     j = branch[ij, T_BUS].real.astype(np.int64)
    #     for m in pq:
    #         mm = pq_lookup[m]
    #         if m == i:
    #             n = j
    #         elif m == j:
    #             n = i
    #         else:
    #             continue
    #
    #         # A_mn = Vm[m] ** 2 - Vm[m] * Vm[n] * np.cos(Va[m] - Va[n])
    #         # J32[ij, mm] = 2 * (g[ij]**2 + b[ij]**2) * C[ij] * A_mn / Vm[m]
    #         J32_old[ij, mm] = - 2 * r_theta_pu[ij_lookup] * g[ij_lookup] * (Vm[m] - Vm[n] * np.cos(Va[m] - Va[n])) * C[ij_lookup]

    mf = np.r_[branch[tdpf_lines[in_pq_f], F_BUS].real.astype(np.int64)]
    mt = np.r_[branch[tdpf_lines[in_pq_t], T_BUS].real.astype(np.int64)]
    nf = np.r_[branch[tdpf_lines[in_pq_f], T_BUS].real.astype(np.int64)]
    nt = np.r_[branch[tdpf_lines[in_pq_t], F_BUS].real.astype(np.int64)]

    for in_pq, m, n in ((in_pq_f, mf, nf), (in_pq_t, mt, nt)):
        pq_j = pq_lookup[m]
        J32[tdpf_lines[in_pq], pq_j] = - 2 * r_theta_pu[in_pq] * g[in_pq] * (Vm[m] - Vm[n] * np.cos(Va[m] - Va[n])) * C[in_pq]

    return J32


def create_J33(branch, tdpf_lines, r_theta_pu, Vm, Va, dg_dT):
    """
     | J11 = dP/dd     J12 = dP/dV     J13 = dP/dT  |
     | (N-1)x(N-1)     (N-1)x(M)       (N-1)x(R)    |
     |                                              |
     | J21 = dQ/dd     J22 = dQ/dV     J23 = dQ/dT  |
     | (M)x(N-1)       (M)x(M)         (M)x(R)      |
     |                                              |
     | J31 = ddT/dd    J32 = ddT/dV    J33 = ddT/dT |
     | (R)x(N-1)       (R)x(M)         (R)x(R)      |

    N = Number of buses
    M = Number of PQ buses
    R = Number temperature-dependent branches

    shape = (len(branch), len(bus))
    branch elements by row, bus elements by column
    :param g:
    :param b:
    :param dg_dT:
    :param Vm:
    :param r_theta_pu:
    :param pvpq_lookup:

    """

    nrow = len(branch)
    J33 = eye(nrow, format="csr", dtype=np.float64)

    # J33 = np.zeros(shape=(nrow, nrow))
    # J33[np.arange(nrow), np.arange(nrow)] = 1

    # k = (np.square(g) + np.square(b)) / g
    # p_loss_pu = i_square_pu / k

    #for mn in range(nrow):
    # for mn in tdpf_lines:
    #     #for ij in range(nrow):
    #     for ij_lookup, ij in enumerate(tdpf_lines):
    #         if mn == ij:
    #             i, j = branch[ij, [F_BUS, T_BUS]].real.astype(np.int64)
    #             # J33[mn, ij] = -(1 + 2 * alpha[ij_lookup] * r_ref[ij_lookup] * g[ij_lookup] * i_square_pu[ij_lookup])
    #             J33[mn, ij] = 1 - r_theta_pu[ij_lookup] * (Vm[i]**2 + Vm[j]**2 - 2*Vm[i]*Vm[j]*np.cos(Va[i]-Va[j])) * dg_dT[ij_lookup]

    #use this instead:
    # for ij_lookup, ij in enumerate(tdpf_lines):
    #    i, j = branch[ij, [F_BUS, T_BUS]].real.astype(np.int64)
    #    # J33[mn, ij] = -(1 + 2 * alpha[ij_lookup] * r_ref[ij_lookup] * g[ij_lookup] * i_square_pu[ij_lookup])
    #    J33[ij, ij] = 1 - r_theta_pu[ij_lookup] * (Vm[i] ** 2 + Vm[j] ** 2 - 2 * Vm[i] * Vm[j] * np.cos(Va[i] - Va[j])) * dg_dT[ij_lookup]

    # vectorized with numpy:
    i = branch[tdpf_lines, F_BUS].real.astype(np.int64)
    j = branch[tdpf_lines, T_BUS].real.astype(np.int64)
    J33[tdpf_lines, tdpf_lines] = 1 - r_theta_pu * (Vm[i] ** 2 + Vm[j] ** 2 - 2 * Vm[i] * Vm[j] * np.cos(Va[i] - Va[j])) * dg_dT

    return J33



# -*- coding: utf-8 -*-

# Copyright 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


from numba import jit
from numpy import conj, zeros, complex128
from scipy.sparse import issparse, csr_matrix as sparse
from pandapower.pypower.dSbus_dV import dSbus_dV_dense


# @jit(Tuple((c16[:], c16[:]))(c16[:], i4[:], i4[:], c16[:], c16[:]), nopython=True, cache=False)
@jit(nopython=True, cache=False)
def dSbus_dV_numba_sparse(Yx, Yp, Yj, V, Vnorm, Ibus): # pragma: no cover
    """Computes partial derivatives of power injection w.r.t. voltage.

    Calculates faster with numba and sparse matrices.

    Input: Ybus in CSR sparse form (Yx = data, Yp = indptr, Yj = indices), V and Vnorm (= V / abs(V))

    OUTPUT: data from CSR form of dS_dVm, dS_dVa
    (index pointer and indices are the same as the ones from Ybus)

    Translation of: dS_dVm = dS_dVm = diagV * conj(Ybus * diagVnorm) + conj(diagIbus) * diagVnorm
                             dS_dVa = 1j * diagV * conj(diagIbus - Ybus * diagV)
    """

    # transform input

    # init buffer vector
    buffer = zeros(len(V), dtype=complex128)
    dS_dVm = Yx.copy()
    dS_dVa = Yx.copy()

    # iterate through sparse matrix
    for r in range(len(Yp) - 1):
        for k in range(Yp[r], Yp[r + 1]):
            # Ibus = Ybus * V
            buffer[r] += Yx[k] * V[Yj[k]]
            # Ybus * diag(Vnorm)
            dS_dVm[k] *= Vnorm[Yj[k]]
            # Ybus * diag(V)
            dS_dVa[k] *= V[Yj[k]]

        Ibus[r] += buffer[r]

        # conj(diagIbus) * diagVnorm
        buffer[r] = conj(buffer[r]) * Vnorm[r]

    for r in range(len(Yp) - 1):
        for k in range(Yp[r], Yp[r + 1]):
            # diag(V) * conj(Ybus * diagVnorm)
            dS_dVm[k] = conj(dS_dVm[k]) * V[r]

            if r == Yj[k]:
                # diagonal elements
                dS_dVa[k] = -Ibus[r] + dS_dVa[k]
                dS_dVm[k] += buffer[r]

            # 1j * diagV * conj(diagIbus - Ybus * diagV)
            dS_dVa[k] = conj(-dS_dVa[k]) * (1j * V[r])

    return dS_dVm, dS_dVa


def dSbus_dV(Ybus, V, I=None):
    """
    Calls functions to calculate dS/dV depending on whether Ybus is sparse or not
    """

    if issparse(Ybus):
        # I is substracted from Y*V,
        # therefore it must be negative for numba version of dSbus_dV if it is not zeros anyways
        I = zeros(len(V), dtype=complex128) if I is None else -I
        # calculates sparse data
        dS_dVm, dS_dVa = dSbus_dV_numba_sparse(Ybus.data, Ybus.indptr, Ybus.indices, V, V / abs(V), I)
        # generate sparse CSR matrices with computed data and return them
        return sparse((dS_dVm, Ybus.indices, Ybus.indptr)), sparse((dS_dVa, Ybus.indices, Ybus.indptr))
    else:
        I = zeros(len(V), dtype=complex128) if I is None else I
        return dSbus_dV_dense(Ybus, V, I)



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.
import sys

import geojson.utils
import numpy as np

try:
    from matplotlib.patches import RegularPolygon, Arc, Circle, Rectangle, Ellipse, PathPatch
    from matplotlib.textpath import TextPath
    from matplotlib.transforms import Affine2D

    MATPLOTLIB_INSTALLED = True
except ImportError:
    MATPLOTLIB_INSTALLED = False
from pandapower.auxiliary import soft_dependency_error
from pandapower.plotting.plotting_toolbox import _rotate_dim2, get_color_list, get_angle_list, \
    get_linewidth_list

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def node_patches(node_coords, size, patch_type, colors=None, **kwargs):
    """
    Creates node patches from coordinates translating the patch type into patches.

    :param node_coords: coordinates of the nodes to draw
    :type node_coords: iterable
    :param size: size of the patch (can be interpreted differently, depending on the patch type)
    :type size: float
    :param patch_type: type of patches to create  - can be one of
        - "circle" or "ellipse" for an ellipse (cirlces are just ellipses with the same width \
            + height)\
        - "rect" or "rectangle" for a rectangle\
        - "poly<n>" for a polygon with n edges
    :type patch_type: str
    :param colors: colors or color of the patches
    :type colors: iterable, float
    :param kwargs: additional keyword arguments to pass to the patch initialization \
        (might contain "width", "height", "angle" depending on the patch type)
    :type kwargs: dict
    :return: patches - list of rectangle patches for the nodes
    """
    if patch_type.lower() == 'ellipse' or patch_type.lower() == 'circle':
        # circles are just ellipses
        if patch_type.lower() == "circle" and len(set(kwargs.keys()) & {"width", "height"}) == 1:
            wh = kwargs["width"] if "width" in kwargs else kwargs["height"]
            width = wh
            height = wh
        else:
            width = kwargs.pop("width", 2 * size)
            height = kwargs.pop("height", 2 * size)
        angle = kwargs.pop('angle', 0)
        return ellipse_patches(node_coords, width, height, angle, color=colors, **kwargs)
    elif patch_type.lower() == "rect" or patch_type.lower() == "rectangle":
        width = kwargs.pop("width", 2 * size)
        height = kwargs.pop("height", 2 * size)
        return rectangle_patches(node_coords, width, height, color=colors, **kwargs)
    elif patch_type.lower().startswith("poly"):
        edges = int(patch_type[4:])
        return polygon_patches(node_coords, size, edges, color=colors, **kwargs)
    else:
        logger.error("Wrong patchtype. Please choose a correct patch type.")
        raise ValueError("Wrong patchtype")


def ellipse_patches(node_coords, width, height, angle=0, color=None, **kwargs):
    """
    Function to create a list of ellipse patches from node coordinates.

    :param node_coords: coordinates of the nodes to draw
    :type node_coords: iterable
    :param width: width of the ellipse (described by an exterior rectangle)
    :type width: float
    :param height: height of the ellipse (described by an exterior rectangle)
    :type height: float
    :param angle: angle by which to rotate the ellipse
    :type angle: float
    :param color: color or colors of the patches
    :type color: iterable, float
    :param kwargs: additional keyword arguments to pass to the Ellipse initialization
    :type kwargs: dict
    :return: patches - list of ellipse patches for the nodes
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    patches = list()
    angles = get_angle_list(angle, len(node_coords))
    if color is not None:
        colors = get_color_list(color, len(node_coords))
        for (x, y), col, ang in zip(node_coords, colors, angles):
            patches.append(Ellipse((x, y), width, height, angle=ang, color=col, **kwargs))
    else:
        for (x, y), ang in zip(node_coords, angles):
            patches.append(Ellipse((x, y), width, height, angle=ang, **kwargs))
    return patches


def rectangle_patches(node_coords, width, height, color=None, **kwargs):
    """
    Function to create a list of rectangle patches from node coordinates.

    :param node_coords: coordinates of the nodes to draw
    :type node_coords: iterable
    :param width: width of the rectangle
    :type width: float
    :param height: height of the rectangle
    :type height: float
    :param color: color or colors of the patches
    :type color: iterable, float
    :param kwargs: additional keyword arguments to pass to the Rectangle initialization
    :type kwargs: dict
    :return: patches - list of rectangle patches for the nodes
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    patches = list()
    if color is not None:
        colors = get_color_list(color, len(node_coords))
        for (x, y), col in zip(node_coords, colors):
            patches.append(Rectangle((x - width / 2, y - height / 2), width, height, color=color,
                                     **kwargs))
    else:
        for x, y in node_coords:
            patches.append(Rectangle((x - width / 2, y - height / 2), width, height, **kwargs))
    return patches


def polygon_patches(node_coords, radius, num_edges, color=None, **kwargs):
    """
    Function to create a list of polygon patches from node coordinates. The number of edges for the
    polygon can be defined.

    :param node_coords: coordinates of the nodes to draw
    :type node_coords: iterable
    :param radius: radius for the polygon (from centroid to edges)
    :type radius: float
    :param num_edges: number of edges of the polygon
    :type num_edges: int
    :param color: color or colors of the patches
    :type color: iterable, float
    :param kwargs: additional keyword arguments to pass to the Polygon initialization
    :type kwargs: dict
    :return: patches - list of rectangle patches for the nodes
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    patches = list()
    if color is not None:
        colors = get_color_list(color, len(node_coords))
        for (x, y), col in zip(node_coords, colors):
            patches.append(RegularPolygon([x, y], numVertices=num_edges, radius=radius, color=col,
                                          **kwargs))
    else:
        for x, y in node_coords:
            patches.append(RegularPolygon([x, y], numVertices=num_edges, radius=radius, **kwargs))
    return patches


def load_patches(node_coords, size, angles, **kwargs):
    """
    Creation function of patches for loads.

    :param node_coords: coordinates of the nodes that the loads belong to.
    :type node_coords: iterable
    :param size: size of the patch
    :type size: float
    :param angles: angles by which to rotate the patches (in radians)
    :type angles: iterable(float), float
    :param kwargs: additional keyword arguments (might contain parameters "offset",\
        "patch_edgecolor" and "patch_facecolor")
    :type kwargs:
    :return: Return values are: \
        - lines (list) - list of coordinates for lines leading to load patches\
        - polys (list of RegularPolygon) - list containing the load patches\
        - keywords (set) - set of keywords removed from kwargs
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    offset = kwargs.get("offset", 1.2 * size)
    all_angles = get_angle_list(angles, len(node_coords))
    edgecolor = kwargs.get("patch_edgecolor", "w")
    facecolor = kwargs.get("patch_facecolor", "w")
    edgecolors = get_color_list(edgecolor, len(node_coords))
    facecolors = get_color_list(facecolor, len(node_coords))
    polys, lines = list(), list()
    for i, node_geo in enumerate(node_coords):
        p2 = node_geo + _rotate_dim2(np.array([0, offset + size]), all_angles[i])
        p3 = node_geo + _rotate_dim2(np.array([0, offset + size / 2]), all_angles[i])
        polys.append(RegularPolygon(p2, numVertices=3, radius=size, orientation=-all_angles[i],
                                    fc=facecolors[i], ec=edgecolors[i]))
        lines.append((node_geo, p3))
    return lines, polys, {"offset", "patch_edgecolor", "patch_facecolor"}


def gen_patches(node_coords, size, angles, **kwargs):
    """
    Creation function of patches for generators.

    :param node_coords: coordinates of the nodes that the generators belong to.
    :type node_coords: iterable
    :param size: size of the patch
    :type size: float
    :param angles: angles by which to rotate the patches (in radians)
    :type angles: iterable(float), float
    :param kwargs: additional keyword arguments (might contain parameters "offset",\
        "patch_edgecolor" and "patch_facecolor")
    :type kwargs:
    :return: Return values are: \
        - lines (list) - list of coordinates for lines leading to generator patches\
        - polys (list of RegularPolygon) - list containing the generator patches\
        - keywords (set) - set of keywords removed from kwargs
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    polys, lines = list(), list()
    offset = kwargs.get("offset", 2. * size)
    all_angles = get_angle_list(angles, len(node_coords))
    edgecolor = kwargs.get("patch_edgecolor", "k")
    facecolor = kwargs.get("patch_facecolor", (1, 0, 0, 0))
    edgecolors = get_color_list(edgecolor, len(node_coords))
    facecolors = get_color_list(facecolor, len(node_coords))
    for i, node_geo in enumerate(node_coords):
        p2 = node_geo + _rotate_dim2(np.array([0, size + offset]), all_angles[i])
        polys.append(Circle(p2, size, fc=facecolors[i], ec=edgecolors[i]))
        polys.append(
            Arc(p2 + np.array([-size / 6.2, -size / 2.6]), size / 2, size, theta1=65, theta2=120,
                ec=edgecolors[i]))
        polys.append(
            Arc(p2 + np.array([size / 6.2, size / 2.6]), size / 2, size, theta1=245, theta2=300,
                ec=edgecolors[i]))
        lines.append((node_geo, p2 + _rotate_dim2(np.array([0, size]), -all_angles[i])))
    return lines, polys, {"offset", "patch_edgecolor", "patch_facecolor"}


def sgen_patches(node_coords, size, angles, **kwargs):
    """
    Creation function of patches for static generators.

    :param node_coords: coordinates of the nodes that the static generators belong to.
    :type node_coords: iterable
    :param size: size of the patch
    :type size: float
    :param angles: angles by which to rotate the patches (in radians)
    :type angles: iterable(float), float
    :param kwargs: additional keyword arguments (might contain parameters "offset", "r_triangle",\
        "patch_edgecolor" and "patch_facecolor")
    :type kwargs:
    :return: Return values are: \
        - lines (list) - list of coordinates for lines leading to static generator patches\
        - polys (list of RegularPolygon) - list containing the static generator patches\
        - keywords (set) - set of keywords removed from kwargs
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    polys, lines = list(), list()
    offset = kwargs.get("offset", 2 * size)
    r_triangle = kwargs.get("r_triangles", size * 0.4)
    edgecolor = kwargs.get("patch_edgecolor", "w")
    facecolor = kwargs.get("patch_facecolor", "w")
    edgecolors = get_color_list(edgecolor, len(node_coords))
    facecolors = get_color_list(facecolor, len(node_coords))
    for i, node_geo in enumerate(node_coords):
        mid_circ = node_geo + _rotate_dim2(np.array([0, offset + size]), angles[i])
        circ_edge = node_geo + _rotate_dim2(np.array([0, offset]), angles[i])
        mid_tri1 = mid_circ + _rotate_dim2(np.array([r_triangle, -r_triangle / 4]), angles[i])
        mid_tri2 = mid_circ + _rotate_dim2(np.array([-r_triangle, r_triangle / 4]), angles[i])
        # dropped perpendicular foot of triangle1
        perp_foot1 = mid_tri1 + _rotate_dim2(np.array([0, -r_triangle / 2]), angles[i])
        line_end1 = perp_foot1 + + _rotate_dim2(np.array([-2.5 * r_triangle, 0]), angles[i])
        perp_foot2 = mid_tri2 + _rotate_dim2(np.array([0, r_triangle / 2]), angles[i])
        line_end2 = perp_foot2 + + _rotate_dim2(np.array([2.5 * r_triangle, 0]), angles[i])
        polys.append(Circle(mid_circ, size, fc=facecolors[i], ec=edgecolors[i]))
        polys.append(RegularPolygon(mid_tri1, numVertices=3, radius=r_triangle,
                                    orientation=-angles[i], fc=facecolors[i], ec=edgecolors[i]))
        polys.append(RegularPolygon(mid_tri2, numVertices=3, radius=r_triangle,
                                    orientation=np.pi - angles[i], fc=facecolors[i],
                                    ec=edgecolors[i]))
        lines.append((node_geo, circ_edge))
        lines.append((perp_foot1, line_end1))
        lines.append((perp_foot2, line_end2))
    return lines, polys, {"offset", "r_triangle", "patch_edgecolor", "patch_facecolor"}


def storage_patches(node_coords, size, angles, **kwargs):
    """
    Creation function of patches for storage systems.

    :param node_coords: coordinates of the nodes that the storage system belong to.
    :type node_coords: iterable
    :param size: size of the patch
    :type size: float
    :param angles: angles by which to rotate the patches (in radians)
    :type angles: iterable(float), float
    :param kwargs: additional keyword arguments (might contain parameters "offset", "r_triangle",\
        "patch_edgecolor" and "patch_facecolor")
    :type kwargs:
    :return: Return values are: \
        - lines (list) - list of coordinates for lines leading to storage patches\
        - polys (list of RegularPolygon) - list containing the storage patches\
        - keywords (set) - set of keywords removed from kwargs
    """
    polys, lines = list(), list()
    offset = kwargs.get("offset", 1 * size)
    r_triangle = kwargs.get("r_triangles", size * 0.4)
    for i, node_geo in enumerate(node_coords):
        mid_circ = node_geo + _rotate_dim2(np.array([0, offset + r_triangle * 2.]), angles[i])
        circ_edge = node_geo + _rotate_dim2(np.array([0, offset]), angles[i])
        mid_tri1 = mid_circ + _rotate_dim2(np.array([-r_triangle, -r_triangle]), angles[i])

        # dropped perpendicular foot of triangle1
        perp_foot1 = mid_tri1 + _rotate_dim2(np.array([r_triangle * 0.5, -r_triangle / 4]), angles[i])
        line_end1 = perp_foot1 + _rotate_dim2(np.array([1 * r_triangle, 0]), angles[i])

        perp_foot2 = mid_tri1 + _rotate_dim2(np.array([0, -r_triangle]), angles[i])
        line_end2 = perp_foot2 + _rotate_dim2(np.array([2. * r_triangle, 0]), angles[i])

        lines.append((node_geo, circ_edge))
        lines.append((perp_foot1, line_end1))
        lines.append((perp_foot2, line_end2))
    return lines, polys, {"offset", "r_triangle", "patch_edgecolor", "patch_facecolor"}


def ext_grid_patches(node_coords, size, angles, **kwargs):
    """
    Creation function of patches for external grids.

    :param node_coords: coordinates of the nodes that the external grids belong to.
    :type node_coords: iterable
    :param size: size of the patch
    :type size: float
    :param angles: angles by which to rotate the patches (in radians)
    :type angles: iterable(float), float
    :param kwargs: additional keyword arguments (might contain parameters "offset",\
        "patch_edgecolor" and "patch_facecolor")
    :type kwargs:
    :return: Return values are: \
        - lines (list) - list of coordinates for lines leading to external grid patches\
        - polys (list of RegularPolygon) - list containing the external grid patches\
        - keywords (set) - set of keywords removed from kwargs (empty
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    offset = kwargs.get("offset", 2 * size)
    all_angles = get_angle_list(angles, len(node_coords))
    edgecolor = kwargs.get("patch_edgecolor", "w")
    facecolor = kwargs.get("patch_facecolor", "w")
    edgecolors = get_color_list(edgecolor, len(node_coords))
    facecolors = get_color_list(facecolor, len(node_coords))
    polys, lines = list(), list()
    for i, node_geo in enumerate(node_coords):
        p2 = node_geo + _rotate_dim2(np.array([0, offset]), all_angles[i])
        p_ll = p2 + _rotate_dim2(np.array([-size, 0]), all_angles[i])
        polys.append(Rectangle(p_ll, 2 * size, 2 * size, angle=(-all_angles[i] / np.pi * 180),
                               fc=facecolors[i], ec=edgecolors[i], hatch="XXX"))
        lines.append((node_geo, p2))
    return lines, polys, {"offset", "patch_edgecolor", "patch_facecolor"}


def trafo_patches(coords, size, **kwargs):
    """
    Creates a list of patches and line coordinates representing transformers each connecting two
    nodes.

    :param coords: list of connecting node coordinates (usually should be \
        `[((x11, y11), (x12, y12)), ((x21, y21), (x22, y22)), ...]`)
    :type coords: (N, (2, 2)) shaped iterable
    :param size: size of the trafo patches
    :type size: float
    :param kwargs: additional keyword arguments (might contain parameters "patch_edgecolor" and\
        "patch_facecolor")
    :type kwargs:
    :return: Return values are: \
        - lines (list) - list of coordinates for lines connecting nodes and transformer patches\
        - circles (list of Circle) - list containing the transformer patches (rings)
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    edgecolor = kwargs.get("patch_edgecolor", "w")
    facecolor = kwargs.get("patch_facecolor", (1, 0, 0, 0))
    edgecolors = get_color_list(edgecolor, len(coords))
    facecolors = get_color_list(facecolor, len(coords))
    linewidths = kwargs.get("linewidths", 2.)
    linewidths = get_linewidth_list(linewidths, len(coords), name_entries="trafos")
    circles, lines = list(), list()

    # load and extract the coords from the geojson object
    coords = list(map(lambda x: list(geojson.utils.coords(geojson.loads(x))), coords))

    for i, (p1, p2) in enumerate(coords):
        p1 = np.array(p1)
        p2 = np.array(p2)
        if np.all(p1 == p2):
            continue
        d = np.sqrt(np.sum((p1 - p2) ** 2))
        if size is None:
            size_this = np.sqrt(d) / 5
        else:
            size_this = size
        off = size_this * 0.35
        circ1 = (0.5 - off / d) * (p1 - p2) + p2
        circ2 = (0.5 + off / d) * (p1 - p2) + p2
        circles.append(Circle(circ1, size_this, fc=facecolors[i], ec=edgecolors[i],
                              lw=linewidths[i]))
        circles.append(Circle(circ2, size_this, fc=facecolors[i], ec=edgecolors[i],
                              lw=linewidths[i]))

        lp1 = (0.5 - off / d - size_this / d) * (p2 - p1) + p1
        lp2 = (0.5 - off / d - size_this / d) * (p1 - p2) + p2
        lines.append([p1, lp1])
        lines.append([p2, lp2])
    return lines, circles, {"patch_edgecolor", "patch_facecolor"}


def ward_patches(node_coords, size, angles, **kwargs):
    """
    Creation function of patches for wards.

    :param node_coords: coordinates of the nodes that the wards belong to.
    :type node_coords: iterable
    :param size: size of the patch
    :type size: float
    :param angles: angles by which to rotate the patches (in radians)
    :type angles: iterable(float), float
    :param kwargs: additional keyword arguments (might contain parameters "offset",\
        "patch_edgecolor" and "patch_facecolor")
    :type kwargs:
    :return: Return values are: \
        - lines (list) - list of coordinates for lines leading to ward patches\
        - polys (list of RegularPolygon) - list containing the ward patches\
        - keywords (set) - set of keywords removed from kwargs (empty)
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    offset = kwargs.get("offset", 2 * size)
    all_angles = get_angle_list(angles, len(node_coords))
    edgecolor = kwargs.get("patch_edgecolor", "w")
    facecolor = kwargs.get("patch_facecolor", "w")
    edgecolors = get_color_list(edgecolor, len(node_coords))
    facecolors = get_color_list(facecolor, len(node_coords))
    polys, lines = list(), list()
    for i, node_geo in enumerate(node_coords):
        p2 = node_geo + _rotate_dim2(np.array([0, offset]), all_angles[i])
        p_ll = p2 + _rotate_dim2(np.array([-size, 0]), all_angles[i])
        polys.append(Rectangle(p_ll, 2 * size, 2 * size, angle=(-all_angles[i] / np.pi * 180),
                               fc=facecolors[i], ec=edgecolors[i]))
        lines.append((node_geo, p2))

        text_foot = p2 + np.array([-1.425 * size, -0.34 * size]) + _rotate_dim2(np.array([size, size]), all_angles[i])

        text_path = TextPath((0, 0), "W", size=0.9 * size, prop=dict(weight='light'))
        transform = Affine2D().translate(text_foot[0], text_foot[1]).rotate(all_angles[i])
        text_patch = PathPatch(transform.transform_path(text_path), edgecolor='black', facecolor='black', lw=0.5)
        polys.append(text_patch)

    return lines, polys, {"offset", "patch_edgecolor", "patch_facecolor"}


def xward_patches(node_coords, size, angles, **kwargs):
    """
    Creation function of patches for xwards.

    :param node_coords: coordinates of the nodes that the xwards belong to.
    :type node_coords: iterable
    :param size: size of the patch
    :type size: float
    :param angles: angles by which to rotate the patches (in radians)
    :type angles: iterable(float), float
    :param kwargs: additional keyword arguments (might contain parameters "offset",\
        "patch_edgecolor" and "patch_facecolor")
    :type kwargs:
    :return: Return values are: \
        - lines (list) - list of coordinates for lines leading to elements of xward patches\
        - polys (list of RegularPolygon) - list containing the xward patches\
        - keywords (set) - set of keywords removed from kwargs (empty)
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
    offset = kwargs.get("offset", 2 * size)
    all_angles = get_angle_list(angles, len(node_coords))
    edgecolor = kwargs.get("patch_edgecolor", "k")
    facecolor = kwargs.get("patch_facecolor", "w")
    edgecolors = get_color_list(edgecolor, len(node_coords))
    facecolors = get_color_list(facecolor, len(node_coords))
    polys, lines = list(), list()
    for i, node_geo in enumerate(node_coords):
        p2 = node_geo + _rotate_dim2(np.array([0, offset]), all_angles[i])
        p_ll = p2 + _rotate_dim2(np.array([-size, 0]), all_angles[i])
        polys.append(Rectangle(p_ll, 2 * size, 2 * size, angle=(-all_angles[i] / np.pi * 180),
                               fc=facecolors[i], ec=edgecolors[i]))
        lines.append((node_geo, p2))

        text_foot = p2 + np.array([-1.755 * size, -0.34 * size]) + _rotate_dim2(np.array([size, size]), all_angles[i])

        text_path = TextPath((0, 0), "XW", size=0.9 * size, prop=dict(weight='light'))
        transform = Affine2D().translate(text_foot[0], text_foot[1]).rotate(all_angles[i])
        text_patch = PathPatch(transform.transform_path(text_path), edgecolor='black', facecolor='black', lw=0.5)
        polys.append(text_patch)

    return lines, polys, {"offset", "patch_edgecolor", "patch_facecolor"}


def vsc_patches(coords, size, **kwargs):
    """
    Creates a list of patches and line coordinates representing VSCs each connecting an AC and a DC
    node.

    :param coords: list of connecting node coordinates (usually should be \
        `[((x11, y11), (x12, y12)), ((x21, y21), (x22, y22)), ...]`)
    :type coords: (N, (2, 2)) shaped iterable
    :param size: size of the VSC patches
    :type size: float
    :param kwargs: additional keyword arguments (might contain parameters "patch_edgecolor" and\
        "patch_facecolor")
    :type kwargs:
    :return: Return values are: \
        - lines (list) - list of coordinates for lines connecting nodes and VSC patches\
        - squares (list of Rectangle) - list containing the VSC patches (squares)
    """
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name)+"()", "matplotlib")
    edgecolor = kwargs.get("patch_edgecolor", "w")
    facecolor = kwargs.get("patch_facecolor", "w")
    edgecolors = get_color_list(edgecolor, len(coords))
    facecolors = get_color_list(facecolor, len(coords))
    linewidths = kwargs.get("linewidths", 2.)
    linewidths = get_linewidth_list(linewidths, len(coords), name_entries="vscs")
    squares, lines = list(), list()

    # load and extract the coords from the geojson object
    coords = list(map(lambda x: list(geojson.utils.coords(geojson.loads(x))), coords))

    for i, (p1, p2) in enumerate(coords):
        p1 = np.array(p1)
        p2 = np.array(p2)
        if np.all(p1 == p2):
            continue
        d = np.sqrt(np.sum((p1 - p2) ** 2))  # distance
        if size is None:
            size_this = np.sqrt(d) / 3
        else:
            size_this = size
        xy1 = p1 - np.array([size_this, size_this])
        xy2 = p2 - np.array([size_this, size_this])
        squares.append(Rectangle(xy1, size_this * 2, size_this * 2, fc=facecolors[i], ec=edgecolors[i],
                                 lw=linewidths[i], hatch="+++"))
        squares.append(Rectangle(xy2, size_this * 2, size_this * 2, fc=facecolors[i], ec=edgecolors[i],
                                 lw=linewidths[i], hatch="---"))
        lines.append([p1, p2])
    return lines, squares, {"patch_edgecolor", "patch_facecolor"}



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np

from pandapower.auxiliary import ensure_iterability
from pandapower.control.controller.DERController.PQVAreas import BaseArea, PQVArea4110, QVArea4110
from pandapower.control.controller.DERController.QModels import QModel
from pandapower.control.controller.pq_control import PQController

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------------------------------
# DERController
# -------------------------------------------------------------------------------------------------


class DERController(PQController):
    """Flexible controller to model plenty types of distributed energy resource (DER) control
    characteristics, such as

    + const Q
    + cosphi fixed (different types)
    + cosphi(P) curve
    + Q(V)

    and restrict the behavior to defined PQV areas, such as

    + PQVArea4130V1, PQVArea4130V2, PQVArea4130V3
    + PQVArea4120V1, PQVArea4120V2, PQVArea4120V3
    + PQVArea4110
    + PQVArea4105
    + PQAreaSTATCOM
    + PQVAreaPOLYGON (PQAreaPOLYGON, QVAreaPOLYGON)

    .. tip: For the DER controller, an extensive `tutorial <https://github.com/e2nIEE/pandapower/tree/develop/tutorials>`_ is available.

    .. note:: sn_mva of the controlled elements is expected to be the rated power (generation) of
        the elements (called P_{b,installed} in the VDE AR N standards). Scalings and limits are
        usually relative to that (sn_mva) values.

    INPUT:
        **net** (pandapower net)

        **element_index** (int[]) - IDs of the controlled elements

    OPTIONAL:
        **element** (str, "sgen") - element type which is controlled

        **q_model** (object, None) - an q_model, such as provided in this file, should be passed to
        model how the q value should be determined.

        **pqv_area** (object, None) - an pqv_area, such as provided in this file, should be passed
        to model q values are allowed.

        **saturate_sn_mva** (float, NaN) - Maximum apparent power of the inverter. If given, the
        p or q values (depending on q_prio) are reduced to this maximum apparent power. Usually,
        it is not necessary to pass this values since the inverter needs to be dimensioned to provide
        the standardized reactive power requirements.

        **q_prio** (bool, True) - If True, the active power is reduced first in case of power
        reduction due to saturate_sn_mva. Otherwise, the reactive power is reduced first.

        **damping_coef** (float, 2) - damping coefficient to influence the power updating process
        of the control loop. A higher value mean slower changes of p and q towards the latest target
        values

        **max_p_error** (float, 0.0001) - Maximum absolute error of active power in MW

        **max_q_error** (float, 0.0001) - Maximum absolute error of reactive power in Mvar

        **pq_simultaneity_factor** (float, 1.0) - Simultaneity factor applied to P and Q

        **data_source** ( , None) - A DataSource that contains profiles

        **p_profile** (str[], None) - The profile names of the controlled elements in the data
        source for active power time series values

        **profile_from_name** (bool, False) - If True, the profile names of the controlled elements
        in the data source for active power time series values will be set be the name of the
        controlled elements, e.g. for controlled sgen "SGEN_1", the active power profile "P_SGEN_1"
        is applied

        **profile_scale** (float, 1.0) - A scaling factor applied to the values of profiles

        **in_service** (bool, True) - Indicates if the controller is currently in_service

        **ts_absolute** (bool, True) - Whether the time step values are absolute power values or
        scaling factors

    Example
    -------
    >>> from pandapower.networks.cigre_networks import create_cigre_network_mv
    >>> from pandapower.control.controller.DERController import DERController, QModelCosphiP, PQVArea4120V2
    >>> from pandapower.run import runpp
    >>> net = create_cigre_network_mv(with_der=True)
    >>> controlled_sgens = DERController(
    ...     net, net.sgen.index,
    ...     q_model=QModelCosphiP(cosphi=-0.95),
    ...     pqv_area=PQVArea4120V2()
    ... )
    >>> runpp(net, run_control=True)
    """

    def __init__(self, net, element_index, element="sgen",
                 q_model=None, pqv_area=None,
                 saturate_sn_mva=np.nan, q_prio=True, damping_coef=2,
                 max_p_error=1e-6, max_q_error=1e-6, pq_simultaneity_factor=1., f_sizing=1.,
                 data_source=None, p_profile=None, profile_from_name=False,
                 profile_scale=1.0, in_service=True, ts_absolute=True,
                 order=0, level=0, drop_same_existing_ctrl=False, matching_params=None, **kwargs):
        element_index = list(ensure_iterability(element_index))
        if matching_params is None:
            matching_params = {"element_index": element_index}
        super().__init__(net, element_index=element_index, element=element, max_p_error=max_p_error,
                         max_q_error=max_q_error, pq_simultaneity_factor=pq_simultaneity_factor,
                         f_sizing=f_sizing, data_source=data_source,
                         profile_scale=profile_scale, in_service=in_service,
                         ts_absolute=ts_absolute, initial_run=True,
                         drop_same_existing_ctrl=drop_same_existing_ctrl,
                         matching_params=matching_params, initial_powerflow=False,
                         order=order, level=level, **kwargs)

        # --- init DER Model params
        self.q_model = q_model
        self.pqv_area = pqv_area
        self.saturate_sn_mva = saturate_sn_mva
        self.q_prio = q_prio
        self.damping_coef = damping_coef

        if p_profile is not None:
            p_profile = ensure_iterability(p_profile, len(element_index))
        self.set_p_profile(p_profile, profile_from_name)

        # --- log unexpected param values
        if n_nan_sn := sum(self.sn_mva.isnull()):
            logger.error(f"The DERController relates to sn_mva, but for {n_nan_sn} elements "
                         "sn_mva is NaN.")
        if self.saturate_sn_mva <= 0:
            raise ValueError(f"saturate_sn_mva cannot be <= 0 but is {self.saturate_sn_mva}")
        if self.q_model is not None and not isinstance(self.q_model, QModel):
            logger.warning(f"The Q model is expected of type QModel, however {type(self.q_model)} "
                           "is provided.")
        if self.pqv_area is not None and not isinstance(self.pqv_area, BaseArea):
            logger.warning(f"The PQV area is expected of type BaseArea, however "
                           f"{type(self.pqv_area)} is provided.")

    def time_step(self, net, time):
        # get new values from profiles
        self.read_profiles(time)
        self.p_series_mw = self.p_mw
        self.q_series_mvar = self.q_mvar

    #        self.write_to_net(net)

    def is_converged(self, net):

        self._determine_target_powers(net)

        return np.allclose(self.target_q_mvar, self.q_mvar, atol=self.max_q_error) and \
            np.allclose(self.target_p_mw, self.p_mw, atol=self.max_p_error)

    def control_step(self, net):

        if "target_p_mw" not in vars(self) or "target_q_mvar" not in vars(self):
            self._determine_target_powers(net)

        self.p_mw, self.q_mvar = self.target_p_mw, self.target_q_mvar

        self.write_to_net(net)

    def _determine_target_powers(self, net):
        vm_pu = net.res_bus.loc[self.bus, "vm_pu"].set_axis(self.element_index)
        p_series_mw = getattr(self, "p_series_mw", getattr(self, "p_mw", self.sn_mva))
        q_series_mvar = getattr(self, "q_series_mw", self.q_mvar)

        # --- calculate target p and q -------------------------------------------------------------

        if np.any(p_series_mw < 0):
            logger.info("p_series_mw is forced to be greater/equal zero")
            p_series_mw[p_series_mw < 0] = 0.

        # --- First Step: Calculate/Select P, Q
        p_pu = self._step_p(p_series_mw)
        q_pu = self._step_q(p_series_mw=p_series_mw, q_series_mvar=q_series_mvar, vm_pu=vm_pu)

        # --- Second Step: Saturates P, Q according to SnMVA and PQV_AREA
        if self.saturate_sn_mva or (self.pqv_area is not None):
            p_pu, q_pu = self._saturate(p_pu, q_pu, vm_pu)

        # --- Third Step: Convert relative P, Q to p_mw, q_mvar
        target_p_mw, target_q_mvar = p_pu * self.sn_mva, q_pu * self.sn_mva

        # --- Apply target p and q considering the damping factor coefficient ----------------------
        self.target_p_mw = self.p_mw + (target_p_mw - self.p_mw) / self.damping_coef
        self.target_q_mvar = self.q_mvar + (target_q_mvar - self.q_mvar) / self.damping_coef

    def _step_p(self, p_series_mw=None):
        return p_series_mw / self.sn_mva

    def _step_q(self, p_series_mw=None, q_series_mvar=None, vm_pu=None):
        """Q priority: Q setpoint > Q model > Q series"""
        if self.q_model is not None:
            q_pu = self.q_model.step(vm_pu=vm_pu, p_pu=p_series_mw / self.sn_mva)
        else:
            if q_series_mvar is None:
                raise Exception("No Q_model and no q_profile available.")
            q_pu = q_series_mvar / self.sn_mva
        return q_pu

    def _saturate(self, p_pu, q_pu, vm_pu):
        assert p_pu is not None and q_pu is not None

        # Saturation on given pqv_area
        if self.pqv_area is not None:
            in_area = self.pqv_area.in_area(p_pu, q_pu, vm_pu)
            if not all(in_area):
                min_max_q_pu = self.pqv_area.q_flexibility(
                    p_pu=p_pu[~in_area], vm_pu=vm_pu[~in_area])
                q_pu[~in_area] = np.minimum(np.maximum(
                    q_pu[~in_area], min_max_q_pu[:, 0]), min_max_q_pu[:, 1])

        if not np.isnan(self.saturate_sn_mva):
            p_pu, q_pu = self._saturate_sn_mva_step(p_pu, q_pu, vm_pu)
        return p_pu, q_pu

    def _saturate_sn_mva_step(self, p_pu, q_pu, vm_pu):
        # Saturation on SnMVA according to priority mode
        sat_s_pu = self.saturate_sn_mva / self.sn_mva  # sat_s is relative to sn_mva
        to_saturate = p_pu ** 2 + q_pu ** 2 > sat_s_pu ** 2
        if any(to_saturate):
            if self.q_prio:
                if (
                        isinstance(self.pqv_area, PQVArea4110) or isinstance(self.pqv_area, QVArea4110)
                ) and any(
                    (0.95 < vm[to_saturate]) & (vm[to_saturate] < 1.05) &
                    (-0.328684 < q_pu[to_saturate]) & any(q_pu[to_saturate] < 0.328684)
                ):
                    logger.warning(f"Such kind of saturation is performed that is not in line with"
                                   " VDE AR N 4110: p reduction within 0.95 < vm < 1.05 and "
                                   "0.95 < cosphi.")
                q_pu[to_saturate] = np.clip(q_pu[to_saturate], -sat_s_pu[to_saturate],
                                            sat_s_pu[to_saturate])
                p_pu[to_saturate] = np.sqrt(sat_s_pu[to_saturate] ** 2 - q_pu[to_saturate] ** 2)
            else:
                p_pu[to_saturate] = np.clip(p_pu[to_saturate], 0., sat_s_pu[to_saturate])
                q_pu[to_saturate] = np.sqrt(sat_s_pu[to_saturate] ** 2 - p_pu[to_saturate] ** 2) * \
                                    np.sign(q_pu[to_saturate])
        return p_pu, q_pu

    def __str__(self):
        el_id_str = f"len(element_index)={len(self.element_index)}" if len(self.element_index) > 6 \
            else f"element_index={self.element_index}"
        return (f"DERController({el_id_str}, q_model={self.q_model}, pqv_area={self.pqv_area}, "
                f"saturate_sn_mva={self.saturate_sn_mva}, q_prio={self.q_prio}, "
                f"damping_coef={self.damping_coef})")


if __name__ == "__main__":
    pass



import logging
import time

import numpy
import pandas as pd

from pandapower.converter.cim import cim_tools
from pandapower.converter.cim.cim2pp import build_pp_net
from pandapower.converter.cim.other_classes import Report, LogLevel, ReportCode

logger = logging.getLogger('cim.cim2pp.converter_classes.geoCoordinatesFromDLCim16')

sc = cim_tools.get_pp_net_special_columns_dict()


class CoordinatesFromDLCim16:
    def __init__(self, cimConverter: build_pp_net.CimConverter):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cimConverter = cimConverter

    def add_coordinates_from_dl_cim16(self, diagram_name: str = None):
        self.logger.info("Creating the coordinates from CGMES DiagramLayout.")
        time_start = time.time()
        # choose a diagram if it is not given (the first one ascending)
        if diagram_name is None:
            diagram_name = self.cimConverter.cim['dl']['Diagram'].sort_values(by='name')['name'].values[0]
        self.logger.debug("Choosing the geo coordinates from diagram %s" % diagram_name)
        if diagram_name != 'all':
            # reduce the source data to the chosen diagram only
            diagram_rdf_id = \
                self.cimConverter.cim['dl']['Diagram']['rdfId'][
                    self.cimConverter.cim['dl']['Diagram']['name'] == diagram_name].values[0]
            dl_do = self.cimConverter.cim['dl']['DiagramObject'][
                self.cimConverter.cim['dl']['DiagramObject']['Diagram'] == diagram_rdf_id]
            dl_do = dl_do.rename(columns={'rdfId': 'DiagramObject'})
        else:
            dl_do = self.cimConverter.cim['dl']['DiagramObject'].copy()
            dl_do = dl_do.rename(columns={'rdfId': 'DiagramObject'})
        dl_data = pd.merge(dl_do, self.cimConverter.cim['dl']['DiagramObjectPoint'], how='left', on='DiagramObject')
        dl_data = dl_data.drop(columns=['rdfId', 'Diagram', 'DiagramObject'])
        # make sure that the columns 'xPosition' and 'yPosition' are floats
        dl_data['xPosition'] = dl_data['xPosition'].astype(float)
        dl_data['yPosition'] = dl_data['yPosition'].astype(float)
        dl_data['coords_str'] = '['+dl_data['xPosition'].astype(str)+', '+dl_data['yPosition'].astype(str)+']'

        # create coordinates for the different assets bus, line, transformer
        for one_ele in ['bus', 'trafo', 'trafo3w', 'switch', 'ext_grid', 'load', 'sgen', 'gen', 'line', 'dcline',
                        'impedance', 'shunt', 'storage', 'ward', 'xward']:
            one_ele_df = self.cimConverter.net[one_ele][[sc['o_id']]]
            one_ele_df = pd.merge(dl_data, one_ele_df, how='inner', left_on='IdentifiedObject', right_on=sc['o_id'])
            one_ele_df = one_ele_df.sort_values(by=[sc['o_id'], 'sequenceNumber'])
            if one_ele != 'line' and one_ele != 'dcline' and one_ele != 'impedance':
                # only one coordinate for each asset (except line and impedance)
                one_ele_df = one_ele_df.drop_duplicates([sc['o_id']], keep='first')
                one_ele_df['diagram'] = '{"coordinates": ' + one_ele_df["coords_str"] + ', "type": "Point"}'
            else:
                # line strings
                one_ele_df['coords'] = one_ele_df[['xPosition', 'yPosition']].values.tolist()
                one_ele_df['coords'] = one_ele_df[['coords']].values.tolist()
                for _, df_group in one_ele_df.groupby(by=sc['o_id']):
                    one_ele_df.at[df_group.index.values[0], 'coords'] = df_group[
                        ['xPosition', 'yPosition']].values.tolist()
                one_ele_df = one_ele_df.drop_duplicates([sc['o_id']], keep='first')
                one_ele_df['coords'] = one_ele_df['coords'].astype(str)
                one_ele_df['diagram'] = '{"coordinates": '+one_ele_df['coords'].astype(str)+', "type": "LineString"}'

            self.cimConverter.net[one_ele]['diagram'] = self.cimConverter.net[one_ele][sc['o_id']].map(
                one_ele_df.set_index(sc['o_id']).to_dict(orient='dict').get('diagram'))

        self.logger.info("Finished creating the DL coordinates, needed time: %ss" % (time.time() - time_start))
        self.cimConverter.report_container.add_log(Report(
            level=LogLevel.INFO, code=ReportCode.INFO_CONVERTING,
            message="Created the DL coordinates, needed time: %ss" % (time.time() - time_start)))



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.
import numbers
import numpy as np
from pandapower.timeseries.data_source import DataSource

try:
    import pandaplan.core.pplog as pplog
except ImportError:
    import logging as pplog

logger = pplog.getLogger(__name__)


class DFData(DataSource):
    """
    Hold a pandas.DataFrame as the data source, that uses index column as time step.
    Take care that the data is numeric and all the time steps are present.

    Please note: The columns should be integers for scalar accessing!

    INPUT:
        **df** - A pandas DataFrame which holds the time series data

    OPTIONAL:

        **multi** (bool, False) - If True casts columns and indexes to integers.
        This might be necessary if you read your data from csv.
    """

    def __init__(self, df, multi=False):
        super().__init__()
        self.df = df
        if multi:
            # casting column and index to int for multi- columns accessing
            self.df.index = self.df.index.astype(np.int64)
            self.df.columns = self.df.columns.astype(np.int64)

    def __repr__(self):
        s = "%s with %d rows and %d columns" % (
            self.__class__.__name__, len(self.df), len(self.df.columns))

        if len(self.df.columns) <= 10:
            s += ": %s" % self.df.columns.values.__str__()
        return s

    def get_time_step_value(self, time_step, profile_name, scale_factor=1.0):
        res = self.df.loc[time_step, profile_name]
        if hasattr(res, 'values'):
            res = res.values
            isnumber = np.issubdtype(res.dtype, np.number)
        else:
            isnumber = isinstance(res, numbers.Number) and not isinstance(res, bool)
        if isnumber:
            res = res * scale_factor
        return res

    def get_time_steps_len(self):
        return len(self.df)



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import uuid
from collections import defaultdict

import numpy as np
import pandas as pd

from pandapower.auxiliary import get_indices
from pandapower.create import create_empty_network
from pandapower.toolbox.comparison import compare_arrays
from pandapower.toolbox.element_selection import element_bus_tuples, pp_elements

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def add_column_from_node_to_elements(net, column, replace, elements=None, branch_bus=None,
                                     verbose=True):
    """
    Adds column data to elements, inferring them from the column data of buses they are
    connected to.

    INPUT:
        **net** (pandapowerNet) - the pandapower net that will be changed

        **column** (string) - name of column that should be copied from the bus table to the element
        table

        **replace** (boolean) - if True, an existing column in the element table will be overwritten

        **elements** (list) - list of elements that should get the column values from the bus table

        **branch_bus** (list) - defines which bus should be considered for branch elements.
        'branch_bus' must have the length of 2. One entry must be 'from_bus' or 'to_bus', the
        other 'hv_bus' or 'lv_bus'

    EXAMPLE:
        compare to add_zones_to_elements()
    """
    branch_bus = ["from_bus", "hv_bus"] if branch_bus is None else branch_bus
    if column not in net.bus.columns:
        raise ValueError("%s is not in net.bus.columns" % column)
    elements = elements if elements is not None else pp_elements(bus=False, other_elements=False)
    elements_to_replace = elements if replace else [
        el for el in elements if column not in net[el].columns or net[el][column].isnull().all()]
    # bus elements
    for element, bus_type in element_bus_tuples(bus_elements=True, branch_elements=False):
        if element in elements_to_replace:
            net[element][column] = net["bus"][column].loc[net[element][bus_type]].values
    # branch elements
    to_validate = {}
    for element, bus_type in element_bus_tuples(bus_elements=False, branch_elements=True):
        if element in elements_to_replace:
            if bus_type in (branch_bus + ["bus"]):  # copy data, append branch_bus for switch.bus
                net[element][column] = net["bus"][column].loc[net[element][bus_type]].values
            else:  # save data for validation
                to_validate[element] = net["bus"][column].loc[net[element][bus_type]].values
    # validate branch elements, but do not validate double and switches at all
    already_validated = ["switch"]
    for element, bus_type in element_bus_tuples(bus_elements=False, branch_elements=True):
        if (element in elements_to_replace) & (element not in already_validated):
            already_validated += [element]
            crossing = sum(~compare_arrays(net[element][column].values, to_validate[element]))
            if crossing > 0:
                if verbose:
                    logger.warning("There have been %i %ss with different " % (crossing, element) +
                                   "%s data at from-/hv- and to-/lv-bus" % column)
                else:
                    logger.debug("There have been %i %ss with different " % (crossing, element) +
                                 "%s data at from-/hv- and to-/lv-bus" % column)


def add_column_from_element_to_elements(net, column, replace, elements=None,
                                        continue_on_missing_column=True):
    """
    Adds column data to elements, inferring them from the column data of the elements linked by the
    columns "element" and "element_type" or "et".

    INPUT:
        **net** (pandapowerNet) - the pandapower net that will be changed

        **column** (string) - name of column that should be copied from the tables of the elements.

        **replace** (boolean) - if True, an existing column will be overwritten

        **elements** (list) - list of elements that should get the column values from the linked
        element tables. If None, all elements with the columns "element" and "element_type" or
        "et" are considered (these are currently "measurement" and "switch").

        **continue_on_missing_column** (Boolean, True) - If False, a error will be raised in case of
        an element table has no column 'column' although this element is refered in 'elements'.
        E.g. 'measurement' is in 'elements' and in net.measurement is a trafo measurement but
        in net.trafo there is no column 'name' although column=='name' - ni this case
        'continue_on_missing_column' acts.

    EXAMPLE:
        >>> from pandapower.create import create_measurement
        >>> from pandapower import add_column_from_element_to_elements
        >>> from pandapower.networks.cigre_networks import create_cigre_network_mv
        >>> net = create_cigre_network_mv()
        >>> create_measurement(net, "i", "trafo", 5, 3, 0, side="hv")
        >>> create_measurement(net, "i", "line", 5, 3, 0, side="to")
        >>> create_measurement(net, "p", "bus", 5, 3, 2)
        >>> print(net.measurement.name.values, net.switch.name.values)
        >>> add_column_from_element_to_elements(net, "name", True)
        >>> print(net.measurement.name.values, net.switch.name.values)
    """
    elements = elements if elements is not None else pp_elements()
    elements_with_el_and_et_column = [el for el in elements if "element" in net[el].columns and (
            "element_type" in net[el].columns or "et" in net[el].columns)]
    elements_to_replace = elements_with_el_and_et_column if replace else [
        el for el in elements_with_el_and_et_column if column not in net[el].columns or net[el][
            column].isnull().all()]
    for el in elements_to_replace:
        et_col = "element_type" if "element_type" in net[el].columns else "et"
        element_type = net[el][et_col]
        for short, complete in [("t", "trafo"), ("t3", "trafo3w"), ("l", "line"), ("s", "switch"),
                                ("b", "bus")]:
            element_type.loc[element_type == short] = complete
        element_types_without_column = [et for et in set(element_type) if column not in
                                        net[et].columns]
        if len(element_types_without_column):
            message = "%s is not in net[et].columns with et in " % column + str(
                element_types_without_column)
            if not continue_on_missing_column:
                raise KeyError(message)
            else:
                logger.debug(message)
        for et in list(set(element_type) - set(element_types_without_column)):
            idx_et = element_type.index[element_type == et]
            net[el].loc[idx_et, column] = net[et][column].loc[net[el].element[idx_et]].values


def add_zones_to_elements(net, replace=True, elements=None, **kwargs):
    """
    Adds zones to elements, inferring them from the zones of buses they are connected to.
    """
    elements = pp_elements(bus=False) if elements is None else elements
    add_column_from_node_to_elements(net, "zone", replace=replace, elements=elements, **kwargs)


def reindex_buses(net, bus_lookup):
    """
    Changes the index of net.bus and considers the new bus indices in all other pandapower element
    tables.

    INPUT:
      **net** - pandapower network

      **bus_lookup** (dict) - the keys are the old bus indices, the values the new bus indices
    """
    not_fitting_bus_lookup_keys = set(bus_lookup.keys()) - set(net.bus.index)
    if len(not_fitting_bus_lookup_keys):
        logger.error("These bus indices are unknown to net. Thus, they cannot be reindexed: " +
                     str(not_fitting_bus_lookup_keys))

    missing_bus_indices = sorted(set(net.bus.index) - set(bus_lookup.keys()))
    if len(missing_bus_indices):
        bus_lookup.update({b: b for b in missing_bus_indices})

    # --- reindex buses
    net.bus.index = get_indices(net.bus.index, bus_lookup)
    net.res_bus.index = get_indices(net.res_bus.index, bus_lookup)
    net.res_bus_3ph.index = get_indices(net.res_bus_3ph.index, bus_lookup)
    net.res_bus_sc.index = get_indices(net.res_bus_sc.index, bus_lookup)

    # --- adapt link in bus elements
    for element, value in element_bus_tuples():
        net[element][value] = get_indices(net[element][value], bus_lookup)
    if "bus_geodata" in net:
        net["bus_geodata"].set_index(get_indices(net["bus_geodata"].index, bus_lookup), inplace=True)

    # --- adapt group link
    if net.group.shape[0]:
        for row in np.arange(net.group.shape[0], dtype=np.int64)[
            (net.group.element_type == "bus").values & net.group.reference_column.isnull().values]:
            net.group.iat[row, net.group.columns.get_loc("element_index")] = list(
                get_indices(net.group.element_index.iat[row], bus_lookup))

    # --- adapt measurement link
    bus_meas = net.measurement.element_type == "bus"
    net.measurement.loc[bus_meas, "element"] = get_indices(net.measurement.loc[bus_meas, "element"],
                                                           bus_lookup)
    side_meas = pd.to_numeric(net.measurement.side, errors="coerce").notnull()
    net.measurement.loc[side_meas, "side"] = get_indices(net.measurement.loc[side_meas, "side"],
                                                         bus_lookup)

    # --- adapt switch link
    bb_switches = net.switch[net.switch.et == "b"]
    net.switch.loc[bb_switches.index, "element"] = get_indices(bb_switches.element, bus_lookup)

    return bus_lookup


def create_continuous_bus_index(net, start=0, store_old_index=False):
    """
    Creates a continuous bus index starting at 'start' and replaces all
    references of old indices by the new ones.

    INPUT:
      **net** - pandapower network

    OPTIONAL:
      **start** - index begins with "start"

      **store_old_index** - if True, stores the old index in net.bus["old_index"]

    OUTPUT:
      **bus_lookup** - mapping of old to new index
    """
    net.bus.sort_index(inplace=True)
    if store_old_index:
        net.bus["old_index"] = net.bus.index.values
    new_bus_idxs = list(np.arange(start, len(net.bus) + start))
    bus_lookup = dict(zip(net["bus"].index.values, new_bus_idxs))
    reindex_buses(net, bus_lookup)
    return bus_lookup


def reindex_elements(net, element_type, new_indices=None, old_indices=None, lookup=None):
    """
    Changes the index of the DataFrame net[element_type].

    Parameters
    ----------
    net : pp.pandapowerNet
        net with elements to reindex
    element_type : str
        name of element type to rename, e.g. "gen" or "load"
    new_indices : typing.Union[list[int], pandas.Index[int]], optional
        new indices to set, by default None
    old_indices : typing.Union[list[int], pandas.Index[int]], optional
        old indices to be replaced. If not given, all indices are
        assumed in case of given new_indices, and all lookup keys are assumed in case of given
        lookup, by default None
    lookup : dict[int,int], optional
        lookup to assign new indices to old indices, by default None

    Notes
    -----
    Either new_indices or lookup must be given.
    old_indices can be given to limit the indices to be replaced. In case of given new_indices,
    both must have the same length.
    If element_type is "group", be careful to give new_indices without passing old_indices because
    group indices do not need to be unique.

    Examples
    --------
    >>> net = pp.create_empty_network()
    >>> idx0 = pp.create_bus(net, 110)
    >>> idx1 = 4
    >>> idx2 = 7
    >>> # Reindex using 'new_indices':
    >>> pp.reindex_elements(net, "bus", [idx1])  # passing old_indices=[idx0] is optional
    >>> net.bus.index
    Int64Index([4], dtype='int64')
    >>> # Reindex using 'lookup':
    >>> pp.reindex_elements(net, "bus", lookup={idx1: idx2})
    Int64Index([7], dtype='int64')
    """
    if not net[element_type].shape[0]:
        return
    if new_indices is None and lookup is None:
        raise ValueError("Either new_indices or lookup must be given.")
    elif new_indices is not None and lookup is not None:
        raise ValueError("Only one can be considered, new_indices or lookup.")
    if new_indices is not None and not len(new_indices) or lookup is not None and not len(
            lookup.keys()):
        return

    if new_indices is not None:
        old_indices = old_indices if old_indices is not None else net[element_type].index
        assert len(new_indices) == len(old_indices)
        lookup = dict(zip(old_indices, new_indices))
    elif old_indices is None:
        old_indices = net[element_type].index.intersection(lookup.keys())

    if element_type == "bus":
        reindex_buses(net, lookup)
        return

    # --- reindex
    new_index = pd.Series(net[element_type].index, index=net[element_type].index)
    if element_type != "group":
        new_index.loc[old_indices] = get_indices(old_indices, lookup)
    else:
        new_index.loc[old_indices] = get_indices(new_index.loc[old_indices].values, lookup)
    net[element_type].set_index(pd.Index(new_index.values), inplace=True)

    # --- adapt group link
    if net.group.shape[0]:
        for row in np.arange(net.group.shape[0], dtype=np.int64)[
            (net.group.element_type == element_type).values & \
            net.group.reference_column.isnull().values]:
            net.group.iat[row, net.group.columns.get_loc("element_index")] = list(
                get_indices(net.group.element_index.iat[row], lookup))

    # --- adapt measurement link
    if element_type in ["line", "trafo", "trafo3w"]:
        affected = net.measurement[(net.measurement.element_type == element_type) &
                                   (net.measurement.element.isin(old_indices))]
        if len(affected):
            net.measurement.loc[affected.index, "element"] = get_indices(affected.element, lookup)

    # --- adapt switch link
    if element_type in ["line", "trafo"]:
        affected = net.switch[(net.switch.et == element_type[0]) &
                              (net.switch.element.isin(old_indices))]
        if len(affected):
            net.switch.loc[affected.index, "element"] = get_indices(affected.element, lookup)

    # --- adapt line_geodata index
    if element_type == "line" and "line_geodata" in net and net["line_geodata"].shape[0]:
        idx_name = net.line_geodata.index.name
        place_holder = uuid.uuid4()
        net["line_geodata"][place_holder] = net["line_geodata"].index
        net["line_geodata"].loc[old_indices.intersection(net.line_geodata.index), place_holder] = (
            get_indices(old_indices.intersection(net.line_geodata.index), lookup))
        net["line_geodata"] = net["line_geodata"].set_index(place_holder)
        net["line_geodata"].index.name = idx_name

    # --- adapt index in cost dataframes
    for cost_df in ["pwl_cost", "poly_cost"]:
        element_in_cost_df = (net[cost_df].et == element_type) & net[cost_df].element.isin(old_indices)
        if sum(element_in_cost_df):
            net[cost_df].loc[element_in_cost_df, "element"] = get_indices(net[cost_df].element[
                                                                              element_in_cost_df], lookup)

    # --- adapt tap_characteristic
    if element_type ==  "trafo":
        if "trafo_characteristic_table" in net and  "id_characteristic" in net["trafo_characteristic_table"]:
            net["trafo_characteristic_table"]["id_characteristic"] = net["trafo_characteristic_table"]["id_characteristic"].map(lookup)

def create_continuous_elements_index(net, start=0, add_df_to_reindex=set()):
    """
    Creating a continuous index for all the elements, starting at zero and replaces all references
    of old indices by the new ones.

    INPUT:
      **net** - pandapower network with unodered indices

    OPTIONAL:
      **start** - index begins with "start"

      **add_df_to_reindex** - by default all useful pandapower elements for power flow will be
      selected. Customized DataFrames can also be considered here.

    OUTPUT:
      **net** - pandapower network with odered and continuous indices

    """
    element_types = pp_elements(res_elements=True)

    # create continuous bus index
    create_continuous_bus_index(net, start=start)
    element_types -= {"bus", "bus_geodata", "res_bus"}

    element_types |= add_df_to_reindex

    # run reindex_elements() for all element_types
    for et in list(element_types):
        net[et].sort_index(inplace=True)
        new_index = list(np.arange(start, len(net[et]) + start))

        if et in net and isinstance(net[et], pd.DataFrame):
            if et in ["bus_geodata", "line_geodata"]:
                logger.info(et + " don't need to be included to 'add_df_to_reindex'. It is " +
                            "already included by et=='" + et.split("_")[0] + "'.")
            else:
                reindex_elements(net, et, new_index)
        else:
            logger.debug("No indices could be changed for element '%s'." % et)


def set_scaling_by_type(net, scalings, scale_load=True, scale_sgen=True):
    """
    Sets scaling of loads and/or sgens according to a dictionary
    mapping type to a scaling factor. Note that the type-string is case
    sensitive.
    E.g. scaling = {"pv": 0.8, "bhkw": 0.6}

    :param net:
    :param scalings: A dictionary containing a mapping from element type to
    :param scale_load:
    :param scale_sgen:
    """
    if not isinstance(scalings, dict):
        raise UserWarning("The parameter scaling has to be a dictionary, "
                          "see docstring")

    def scaleit(what):
        et = net[what]
        et["scaling"] = [scale[t] if scale[t] is not None else s for t, s in
                         zip(et.type.values, et.scaling.values)]

    scale = defaultdict(lambda: None, scalings)
    if scale_load:
        scaleit("load")
    if scale_sgen:
        scaleit("sgen")


def set_data_type_of_columns_to_default(net):
    """
    Overwrites dtype of DataFrame columns of PandapowerNet elements to default dtypes defined in
    pandapower. The function "convert_format" does that authomatically for nets saved with
    pandapower versions below 1.6. If this is required for versions starting with 1.6, it should be
    done manually with this function.

    INPUT:
      **net** - pandapower network with unodered indices

    OUTPUT:
      No output; the net passed as input has pandapower-default dtypes of columns in element tables.

    """
    new_net = create_empty_network()
    for key, item in net.items():
        if isinstance(item, pd.DataFrame):
            for col in item.columns:
                if key in new_net and col in new_net[key].columns:
                    if new_net[key][col].dtype == net[key][col].dtype:
                        continue
                    if set(item.columns) == set(new_net[key]):
                        net[key] = net[key].reindex(new_net[key].columns, axis=1)
                    net[key][col] = net[key][col].astype(new_net[key][col].dtype,
                                                         errors="ignore")



import copy
import numpy as np
import pandas as pd

from pandapower.protection.basic_protection_device import ProtectionDevice
from pandapower.protection.utility_functions import create_sc_bus,bus_path_multiple_ext_bus,get_line_path,parallel_lines
from pandapower.shortcircuit.calc_sc import calc_sc
try:
    import matplotlib.pyplot as plt

    MATPLOTLIB_INSTALLED = True
except ImportError:
    MATPLOTLIB_INSTALLED = False


class OCRelay(ProtectionDevice):
    """
    OC Relay used in circuit protection

    INPUT:
        **net** (attrdict) - pandapower net

        **switch_index** (int) - index of the switch that the relay acts upon

        **oc_relay_type** (str) - string specifying the type of over-current protection. Must be either "DTOC", "IDMT",
        or "IDTOC".

        **time_settings** (list of DataFrame) - If given as a list, the time grading will be calculated based on
        topological grid search, and manual tripping time can be provided as a dataframe by respecting the column names.

                For DTOC:
                time_settings =[t>>, t>, t_diff] or Dataframe columns as 'switch_id', 't_gg', 't_g'

                - t>> (t_gg): instantaneous tripping time in seconds
                - t> (t_g):  primary backup tripping time in seconds,
                - t_diff: time grading delay difference in seconds


                For IDMT:
                time_settings =[tms, t_grade] or Dataframe columns as 'switch_id', 'tms', 't_grade'

                - tms: time multiplier settings in seconds
                - t_grade:  time grading delay difference in seconds

                For IDTOC:
                time_settings =[t>>, t>, t_diff, tms,t_grade] or Dataframe columns as 'switch_id', 't_gg', 't_g','tms',
                 't_grade'

                - t>> (t_gg): instantaneous tripping time in seconds
                - t> (t_g):  primary backup tripping time in seconds,
                - t_diff: time grading delay difference in seconds
                - tms: time multiplier settings in seconds
                - t_grade:  time grading delay difference in seconds

        **overload_factor** - (float, 1.25)- Allowable overloading on the line used to calculate the pick-up current

        **ct_current_factor** -(float, 1.2) - Current multiplication factor to calculate the pick-up current

        **safety_factor** -(float, 1) - Safety limit for the instantaneous pick-up current

        **inverse_overload_factor** -(float, 1.2)- Allowable inverse overloading to define the pick-up current in IDMT
        relay

        **pickup_current_manual** - (DataFrame, None) - User-defined relay trip currents can be given as a dataframe.

                DTOC: Dataframe with columns as 'switch_id', 'I_gg', 'I_g'

                IDMT: Dataframe with columns as 'switch_id', 'I_s'

                IDTOC: Dataframe with columns as 'switch_id', 'I_gg', 'I_g', 'I_s'

        **in_service** (bool, True) - specify whether relay is in service

        **overwrite** (bool, True) - specify whether this oc relay should overwrite existing protection devices acting
        on the switch

        **sc_fraction** (float, 0.95) - Maximum possible extent to which the short circuit can be created on the line

        **curve_type** (str, 'standard_inverse') - specify the type of curve used in IDMT and IDTOC.

            Curve type can be:
                 - 'standard_inverse'
                 - 'very_inverse',
                 - 'extremely_inverse',
                 - 'long_inverse'
    """

    def __init__(self, net, switch_index, oc_relay_type, time_settings, overload_factor=1.2, ct_current_factor=1.25,
                 safety_factor=1, inverse_overload_factor=1.2, pickup_current_manual=None, in_service=True,
                 overwrite=True, sc_fraction=0.95, curve_type='standard_inverse', **kwargs):
        super().__init__(net, in_service=in_service, overwrite=overwrite, **kwargs)
        self.switch_index = switch_index
        self.oc_relay_type = oc_relay_type
        self.time_settings = time_settings
        self.overload_factor = overload_factor
        self.ct_current_factor = ct_current_factor
        self.safety_factor = safety_factor
        self.inverse_overload_factor = inverse_overload_factor
        self.pickup_current_manual = pickup_current_manual
        self.sc_fraction = sc_fraction
        self.curve_type = curve_type
        self.kwargs = kwargs

        self.activation_parameter = "i_ka"
        self.tripped = False
        self.I_g = None
        self.I_gg = None
        self.I_s = None
        self.time_grading = None
        self.t_g = None
        self.t_gg = None
        self.t_grade = None
        self.tms = None
        self.create_protection_function(net=net)

    def create_protection_function(self, net):
        # this function is called when the OCRelay is instantiated to determine the parameters for the
        # protection function
        net_sc = copy.deepcopy(net)

        if (net.switch.closed.at[self.switch_index] == True) and (net.switch.et.at[self.switch_index] == 'l'):
            line_idx = int(net.switch.element.at[self.switch_index])
            net_sc = create_sc_bus(net_sc, line_idx, self.sc_fraction)
            bus_idx = max(net_sc.bus.index)

        elif (net.switch.closed.at[self.switch_index] == True) and (net.switch.et.at[self.switch_index] == "t" or "b"):
            bus_idx = net.switch.bus.at[self.switch_index]
            line_idx = None
        else:
            raise ValueError("OC Relay must be connected to line or transformer element")

        calc_sc(net_sc, bus=bus_idx, branch_results=True)

        if self.oc_relay_type == 'DTOC':
            if self.pickup_current_manual is None:
                self.I_g = float(net_sc.line.max_i_ka.iloc[line_idx]) * self.overload_factor * self.ct_current_factor
                self.I_gg = float(net_sc.res_line_sc.ikss_ka.iloc[line_idx]) * self.safety_factor
            else:
                self.I_g = float(self.pickup_current_manual.I_g.iloc[self.switch_index])
                self.I_gg = float(self.pickup_current_manual.I_g.iloc[self.switch_index])
            self.time_grading = time_grading(net, self.time_settings)
            self.t_g = float(self.time_grading.t_g[self.switch_index])
            self.t_gg = float(self.time_grading.t_gg[self.switch_index])

        if self.oc_relay_type == 'IDMT':
            if self.pickup_current_manual is None:
                self.I_s = float(net_sc.line.max_i_ka.iloc[line_idx]) * self.inverse_overload_factor
            else:
                self.I_s = float(self.pickup_current_manual.I_s.iloc[self.switch_index])
            self._select_k_alpha()
            self.time_grading = time_grading(net, self.time_settings)
            self.t_grade = float(self.time_grading.t_g[self.switch_index])
            self.tms = float(self.time_grading.t_gg[self.switch_index])

        if self.oc_relay_type == 'IDTOC':
            if self.pickup_current_manual is None:
                self.I_g = float(net_sc.line.max_i_ka.iloc[line_idx]) * self.overload_factor * self.ct_current_factor
                self.I_gg = float(net_sc.res_line_sc.ikss_ka.iloc[line_idx]) * self.safety_factor
                self.I_s = float(net_sc.line.max_i_ka.iloc[line_idx]) * self.inverse_overload_factor
            else:
                self.I_g = float(self.pickup_current_manual.I_g.iloc[self.switch_index])
                self.I_gg = float(self.pickup_current_manual.I_g.iloc[self.switch_index])
                self.I_s = float(self.pickup_current_manual.I_s.iloc[self.switch_index])
            self._select_k_alpha()
            # calculate time grading first for DTOC part to obtain t_g and t_gg
            self.time_grading = time_grading(net, [self.time_settings[0], self.time_settings[1], self.time_settings[2]])
            self.t_g = float(self.time_grading.t_g[self.switch_index])
            self.t_gg = float(self.time_grading.t_gg[self.switch_index])
            # calculate time grading again for IDMT part to obtain t_grade and tms
            self.time_grading = time_grading(net, [self.time_settings[3], self.time_settings[4]])
            self.tms = float(self.time_grading.t_gg[self.switch_index])
            self.t_grade = float(self.time_grading.t_g[self.switch_index])

    def _select_k_alpha(self):
        if self.curve_type == 'standard_inverse':
            self.k = 0.140
            self.alpha = 0.02
        elif self.curve_type == 'very_inverse':
            self.k = 13.5
            self.alpha = 1
        elif self.curve_type == 'extremely_inverse':
            self.k = 80
            self.alpha = 2
        elif self.curve_type == 'long_inverse':
            self.k = 120
            self.alpha = 1

    def reset_device(self):
        self.tripped = False

    def has_tripped(self):
        return self.tripped

    def status_to_net(self, net):
        # update self.tripped status to net
        net.switch.at[self.switch_index, "closed"] = not self.tripped

    def protection_function(self, net, scenario):
        # compute protection time in net under short-circuit or operating conditions
        if scenario == "sc":
            i_ka = net.res_switch_sc.ikss_ka.at[self.switch_index]
        elif scenario == "pp":
            i_ka = net.res_switch.i_ka.at[self.switch_index]
        else:
            raise ValueError("scenario must be either sc or pp")

        if self.oc_relay_type == 'DTOC':
            if i_ka > self.I_gg:
                self.tripped = True
                act_time_s = self.t_gg
            elif i_ka > self.I_g:
                self.tripped = True
                act_time_s = self.t_g
            else:
                self.tripped = False
                act_time_s = np.inf

        if self.oc_relay_type == 'IDMT':
            if i_ka > self.I_s:
                self.tripped = True
                act_time_s = (self.tms * self.k) / (((i_ka/self.I_s)**self.alpha)-1) + self.t_grade
            else:
                self.tripped = False
                act_time_s = np.inf

        if self.oc_relay_type == 'IDTOC':
            if i_ka > self.I_gg:
                self.tripped = True
                act_time_s = self.t_gg
            elif i_ka > self.I_g:
                self.tripped = True
                act_time_s = self.t_g
            elif i_ka > self.I_s:
                self.tripped = True
                act_time_s = (self.tms * self.k) / (((i_ka / self.I_s)**self.alpha)-1) + self.t_grade
            else:
                self.tripped = False
                act_time_s = np.inf

        protection_result = {"switch_id": self.switch_index,
                             "protection_type": self.__class__.__name__,
                             "trip_melt": self.has_tripped(),
                             "activation_parameter": self.activation_parameter,
                             "activation_parameter_value": i_ka,
                             "trip_melt_time_s": act_time_s}
        return protection_result

    def plot_protection_characteristic(self, net, num=60, xlabel="I [A]", ylabel="time [s]", xmin=10, xmax=10000,
                                       ymin=0.01, ymax=10000, title="Time-Current Characteristic of OC Relay "):

        if self.oc_relay_type == 'DTOC':
            plt.loglog(0, 0)
            plt.step(np.array([self.I_g*1000, self.I_gg*1000, xmax]), np.array([ymax, self.t_g, self.t_gg]))

        elif self.oc_relay_type == 'IDMT':
            x = np.logspace(np.log10((1000*self.I_s)+0.001), np.log10(xmax), 60)
            plt.loglog(x, (self.tms * self.k) / (((x/(1000*self.I_s))**self.alpha)-1) + self.t_grade)

        elif self.oc_relay_type == 'IDTOC':
            x = np.logspace(np.log10((1000 * self.I_s) + 0.001), np.log10(1000 * self.I_g), num=num)
            plt.loglog(x, (self.tms * self.k) / (((x / (1000 * self.I_s)) ** self.alpha) - 1) + self.t_grade)
            plt.step(np.array([self.I_g * 1000, self.I_gg * 1000, xmax]), np.array([
                (self.tms * self.k) / (((self.I_g / self.I_s) ** self.alpha) - 1) + self.t_grade, self.t_g, self.t_gg]))
        else:
            raise ValueError('oc_relay_type must be DTOC, IDMT, or IDTOC')
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.xlim(xmin, xmax)
        plt.ylim(ymin, ymax)
        plt.title(title + str(self.switch_index))
        plt.grid(True, which="both", ls="-")

    def __str__(self):
        s = 'Protection Device: %s \nType: %s \nIndex: %s' % (self.__class__.__name__, self.oc_relay_type, self.switch_index)
        self.characteristic_index = 1
        return s
def time_grading(net,time_settings):

    # Automated time grading calculation
    if isinstance(time_settings,list):
        if len(time_settings)==3:
            time_settings=time_settings
        if len(time_settings)==2:
            time_settings=[time_settings[0], time_settings[1], time_settings[1]]

        # Get the bus path to each lines from all the ext buses
        bus_paths=bus_path_multiple_ext_bus(net)

        # Get all the line paths
        line_paths=[]
        # get sorted line path (last is the longest line path)
        for bus_path in bus_paths:
            line=get_line_path(net, bus_path)
            line_paths.append(line)
            sorted_line_path = sorted(line_paths, key=len)

        #assign tg based on longest line path
        line_length_time={}
        for line_length in range(0,len(sorted_line_path[-1])) :

            distance_counter=len(sorted_line_path[-1])- line_length
            t_g=time_settings[1] + (line_length)* time_settings[2]
            line_length_time[ distance_counter]=t_g


        # line_time gets line id and time
        line_time={}
        for line in sorted_line_path:
            line_length=len(line)
            for length in line_length_time:
                if line_length==length:
                    line_time[line[-1]]=line_length_time[length]
        parallel_line=parallel_lines(net)

        # find the missing lines in line time (due to parallel lines)
        missing_line=[]
        for line in net.line.index:

            linecheck = []
            for key in line_time:

                if line == key:
                    linecheck.append(True)
                else:
                    linecheck.append(False)

            if any(linecheck):

                pass
            else:
                missing_line.append(line)


        # assign time to parallel line from the original time of line

        for parallel in parallel_line:

            for line in missing_line:

                if parallel[0]==line:

                    if parallel[1] not in line_time:
                        pass

                    else:
                        line_time[line]=line_time[parallel[1]]

                if parallel[1]==line:

                    if parallel[0] not in line_time:
                        pass

                    else:
                        line_time[line]=line_time[parallel[0]]


        # Assign time setting to switches
        time_switches={}
        tg_sw_setting=[]
        for switch in net.switch[net.switch.closed == True].index:
            line_id=net.switch[net.switch.closed == True].element.at[switch]

            tg = line_time[line_id]
            time_switches[switch]=tg
            tgg=time_settings[0]
            tg_sw_setting.append([switch,tg,tgg])

            # if there is multiple time setting for each switch take only the highest time
        protection_time_settings = pd.DataFrame(tg_sw_setting)
        protection_time_settings.columns=["switch_id","t_g","t_gg"]
        protection_time_settings= protection_time_settings.sort_values(by=['switch_id'])
        protection_time_settings=protection_time_settings.reset_index(drop=True)

    # Manual time grading settings
    if isinstance(time_settings, pd.DataFrame):
        protection_time_settings = pd.DataFrame(columns=["switch_id", "t_gg", "t_g"])

        if time_settings.columns.values.tolist() == ['switch_id', 't_gg', 't_g']:
            protection_time_settings['switch_id'] = time_settings['switch_id']
            protection_time_settings['t_g'] = time_settings['t_g']
            protection_time_settings['t_gg'] = time_settings['t_gg']

        if time_settings.columns.values.tolist() == ['switch_id', 'tms', 't_grade']:
            protection_time_settings['switch_id'] = time_settings['switch_id']
            protection_time_settings['t_g'] = time_settings['t_grade']
            protection_time_settings['t_gg'] = time_settings['tms']

        if time_settings.columns.values.tolist() == ['switch_id', 't_gg', 't_gg', 'tms', 't_grade']:
            protection_time_settings['switch_id'] = time_settings['switch_id']
            protection_time_settings['t_g'] = time_settings['t_g']
            protection_time_settings['t_gg'] = time_settings['t_gg']
            protection_time_settings['t_grade'] = time_settings['t_grade']
            protection_time_settings['tms'] = time_settings['tms']

    return protection_time_settings


from pandapower import std_type_exists, load_std_type

try:
    import matplotlib.pyplot as plt
    MATPLOTLIB_INSTALLED = True
except ImportError:
    MATPLOTLIB_INSTALLED = False

import numpy as np
from pandapower.control.util.characteristic import LogSplineCharacteristic
from pandapower.protection.basic_protection_device import ProtectionDevice


class Fuse(ProtectionDevice):
    """
    Fuse used in circuit protection

    INPUT:
        **net** (attrdict) - pandapower net

        **switch_index** (int) - index of the switch that the fuse acts upon

    OPTIONAL:

        **fuse_type** (str, "none") - string of the fuse type to be used. If it is in the standard library, it will
        automatically generate the characteristic curve. Otherwise, it has no effect.
        Example: fuse_type = "Siemens NH-1-100"

        **rated_i_a** (float, 0) - the rated current of the fuse in amps

        **characteristic_index** (int, 0) - index of the corresponding characteristic curve. Automatically generated
        when characteristic curve is created.

        **in_service** (bool, True) - indicates if fuse is currently in service and should be included in protection
        computations

        **overwrite** (bool, False) - indicates if fuse should replace already existing protection device acting upon
        switch

        **curve_select** (int, 0) - specifies which curve should be used as characteristic curve. This only has an
        effect for fuses that contain two different melting curves (t_min and t_total). For fuses that only have one
        characteristic curve (t_avg), curve_select has no effect.

        **z_ohm** (float, 0.0001) - gives the resistance of the fuse in ohms. This is used in calculations for bus-bus
        switches

        **name** (str, None) - name of the fuse. For example, name = "Line 2 Fuse"
    """

    def __init__(self, net, switch_index, fuse_type="none", rated_i_a=0, characteristic_index=None, in_service=True,
                 overwrite=False, curve_select=0, z_ohm=0.0001, name=None, **kwargs):
        super().__init__(net, in_service=in_service, overwrite=overwrite, **kwargs)
        self.switch_index = switch_index
        self.fuse_type = fuse_type
        self.in_service = in_service
        self.name = name
        # create protection characteristic curve from std type
        if std_type_exists(net, fuse_type, element="fuse"):
            fuse_data = load_std_type(net=net, name=fuse_type, element="fuse")
            self.rated_i_a = fuse_data['i_rated_a']
            if fuse_data["t_avg"] != 0:
                self.create_characteristic(net, fuse_data["x_avg"], fuse_data["t_avg"])
            elif fuse_data["t_min"] != 0 and curve_select == 0:
                self.create_characteristic(net, fuse_data["x_min"], fuse_data["t_min"])
            elif fuse_data["t_total"] != 0 and curve_select == 1:
                self.create_characteristic(net, fuse_data["x_total"], fuse_data["t_total"])
            else:
                raise ValueError("curve_select must equal 0 or 1")
        else:
            self.rated_i_a = rated_i_a
            self.characteristic_index = characteristic_index
            self.i_start_a = None
            self.i_stop_a = None

        self.activation_parameter = "i_ka"
        self.tripped = False
        self.z_ohm = z_ohm
        net.switch.at[self.switch_index, 'z_ohm'] = self.z_ohm

    def create_characteristic(self, net, x_values, y_values, interpolator_kind="Pchip", **kwargs):
        """
        Create characteristic curve of fuse melting behavior using Piece Cubic Hermite Interpolating Polynomial (PCHIP)
        from scipy.interpolate
        """
        c = LogSplineCharacteristic(net, x_values=x_values, y_values=y_values, interpolator_kind=interpolator_kind,
                                    **kwargs)
        self.characteristic_index = c.index
        self.i_start_a = min(x_values)
        self.i_stop_a = max(x_values)

    def reset_device(self):
        self.tripped = False

    def has_tripped(self):
        return self.tripped

    def status_to_net(self, net):
        # update self.tripped status to net
        net.switch.at[self.switch_index, "closed"] = not self.tripped

    def protection_function(self, net, scenario="sc"):
        # compute protection time in net under short-circuit or operating conditions
        if scenario == "sc":
            i_ka = net.res_switch_sc.ikss_ka.at[self.switch_index]
        elif scenario == "pp":
            i_ka = net.res_switch.i_ka.at[self.switch_index]
        else:
            raise ValueError("scenario must be either sc or pp")
        c = net.characteristic.at[self.characteristic_index, "object"]
        if i_ka * 1000 < self.i_start_a:
            self.tripped = False
            act_time_s = np.inf
        elif i_ka * 1000 <= self.i_stop_a:
            self.tripped = True
            act_time_s = c(i_ka * 1000)
        else:
            self.tripped = True
            act_time_s = 0

        protection_result = {"switch_id": self.switch_index,
                             "protection_type": self.__class__.__name__,
                             "trip_melt": self.has_tripped(),
                             "activation_parameter": self.activation_parameter,
                             "activation_parameter_value": i_ka,
                             "trip_melt_time_s": act_time_s}

        return protection_result

    def plot_protection_characteristic(self, net, num=35, xlabel="I [A]", ylabel="time [s]",
                                       title="Time-Current Characteristic of Fuse"):
        # plots protection characteristic for fuse
        start = min(net.characteristic.at[self.characteristic_index, "object"].x_vals)
        stop = max(net.characteristic.at[self.characteristic_index, "object"].x_vals)
        c = net.characteristic.at[self.characteristic_index, "object"]
        x = np.logspace(start, stop, num)
        y = c(x)
        plt.loglog(x, y)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.xlim(10 ** np.floor(start), 10 ** np.ceil(stop))
        plt.ylim(10 ** np.floor(min(net.characteristic.at[self.characteristic_index, "object"].y_vals)),
                 10 ** np.floor(max(net.characteristic.at[self.characteristic_index, "object"].y_vals)))
        plt.title(title)
        plt.grid(True, which="both", ls="-")

    def __str__(self):  # display Fuse + name instead of Fuse
        s = 'Protection Device: %s \nType: %s \nName: %s' % (self.__class__.__name__, self.fuse_type, self.name)
        self.characteristic_index = 1
        return s



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import numpy as np
import pandas as pd

from pandapower.results_branch import _get_branch_results, _get_branch_results_3ph
from pandapower.results_bus import _get_bus_results, _get_bus_dc_results, _set_buses_out_of_service, \
    _get_shunt_results, _get_p_q_results, _get_bus_v_results, _get_bus_v_results_3ph, _get_p_q_results_3ph, \
    _get_bus_results_3ph, _get_bus_dc_v_results, _get_p_dc_results, _set_dc_buses_out_of_service
from pandapower.results_gen import _get_gen_results, _get_gen_results_3ph, _get_dc_slack_results

BRANCH_RESULTS_KEYS = ("branch_ikss_f", "branch_ikss_t",
                       "branch_ikss_angle_f", "branch_ikss_angle_t",
                       "branch_pkss_f", "branch_pkss_t",
                       "branch_qkss_f", "branch_qkss_t",
                       "branch_vkss_f", "branch_vkss_t",
                       "branch_vkss_angle_f", "branch_vkss_angle_t",
                       "branch_ip_f", "branch_ip_t",
                       "branch_ith_f", "branch_ith_t")

suffix_mode = {"sc": "sc", "se": "est", "pf_3ph": "3ph"}


def _extract_results(net, ppc):
    _set_buses_out_of_service(ppc)  # for NaN results in net.res_bus for inactive buses
    _set_dc_buses_out_of_service(ppc)  # for NaN results in net.res_bus_dc for inactive buses
    bus_lookup_aranged = _get_aranged_lookup(net)
    bus_dc_lookup_aranged = _get_aranged_lookup(net, "bus_dc")
    _get_bus_v_results(net, ppc)
    _get_bus_dc_v_results(net, ppc)
    bus_pq = _get_p_q_results(net, ppc, bus_lookup_aranged)
    _get_shunt_results(net, ppc, bus_lookup_aranged, bus_pq)
    _get_branch_results(net, ppc, bus_lookup_aranged, bus_pq)
    _get_gen_results(net, ppc, bus_lookup_aranged, bus_pq)
    _get_bus_results(net, ppc, bus_pq)
    bus_p_dc = _get_p_dc_results(net, ppc, bus_dc_lookup_aranged)
    # _get_dc_slack_results(net, ppc, bus_dc_lookup_aranged, bus_p_dc)
    _get_bus_dc_results(net, bus_p_dc)
    if net._options["mode"] == "opf":
        _get_costs(net, ppc)
    else:
        _remove_costs(net)


def _extract_results_3ph(net, ppc0, ppc1, ppc2):
    # reset_results(net, False)
    _set_buses_out_of_service(ppc0)
    _set_buses_out_of_service(ppc1)
    _set_buses_out_of_service(ppc2)
    bus_lookup_aranged = _get_aranged_lookup(net)

    _get_bus_v_results_3ph(net, ppc0, ppc1, ppc2)
    bus_pq = _get_p_q_results_3ph(net, bus_lookup_aranged)
    # _get_shunt_results(net, ppc, bus_lookup_aranged, bus_pq)
    _get_branch_results_3ph(net, ppc0, ppc1, ppc2, bus_lookup_aranged, bus_pq)
    _get_gen_results_3ph(net, ppc0, ppc1, ppc2, bus_lookup_aranged, bus_pq)
    _get_bus_results_3ph(net, bus_pq)


def _extract_results_se(net, ppc):
    _set_buses_out_of_service(ppc)
    bus_lookup_aranged = _get_aranged_lookup(net)
    _get_bus_v_results(net, ppc, suffix="_est")
    bus_pq = np.zeros(shape=(len(net["bus"].index), 2), dtype=float)
    _get_branch_results(net, ppc, bus_lookup_aranged, bus_pq, suffix="_est")


def _get_costs(net, ppc):
    net.res_cost = ppc['obj']


def _remove_costs(net):
    if "res_cost" in net.keys():
        del net["res_cost"]


def _get_aranged_lookup(net, bus_table="bus"):
    # generate bus_lookup net -> consecutive ordering
    if len(net[bus_table]) == 0:
        return np.array([], dtype=np.int64)
    maxBus = max(net[bus_table].index.values)
    bus_lookup_aranged = -np.ones(maxBus + 1, dtype=np.int64)
    bus_lookup_aranged[net[bus_table].index.values] = np.arange(len(net[bus_table].index.values))

    return bus_lookup_aranged


def verify_results(net, mode="pf"):
    elements = get_relevant_elements(mode)
    suffix = suffix_mode.get(mode, None)
    for element in elements:
        res_element, res_empty_element = get_result_tables(element, suffix)

        index_equal = False if res_element not in net else net[element].index.equals(net[res_element].index)
        if not index_equal:
            if net["_options"]["init_results"] and element == "bus":
                # if the indices of bus and res_bus are not equal, but init_results is set, the voltage vector
                # is wrong. A UserWarning is raised in this case. For all other elements the result table is emptied.
                raise UserWarning("index of result table '{}' is not equal to the element table '{}'. The init result"
                                  " option may lead to a non-converged power flow.".format(res_element, element))
            # init result table for
            init_element(net, element)
            if element == "bus":
                net._options["init_vm_pu"] = "auto"
                net._options["init_va_degree"] = "auto"


def get_result_tables(element, suffix=None):
    res_element = "res_" + element
    res_element_with_suffix = res_element if suffix is None else res_element + "_%s" % suffix

    if suffix == suffix_mode.get("se", None):
        # State estimation used default result table
        return res_element_with_suffix, "_empty_%s" % res_element
    else:
        return res_element_with_suffix, "_empty_%s" % res_element_with_suffix


def empty_res_element(net, element, suffix=None):
    res_element, res_empty_element = get_result_tables(element, suffix)
    if res_empty_element in net:
        net[res_element] = net[res_empty_element].copy()
    else:
        net[res_element] = pd.DataFrame(columns=pd.Index([], dtype=object),
                                        index=pd.Index([], dtype=np.int64))


def init_element(net, element, suffix=None):
    res_element, res_empty_element = get_result_tables(element, suffix)
    index = net[element].index
    if len(index):
        # init empty dataframe
        if res_empty_element in net:
            columns = net[res_empty_element].columns
            net[res_element] = pd.DataFrame(np.nan, index=index,
                                            columns=columns, dtype='float')
        else:
            net[res_element] = pd.DataFrame(index=index, dtype='float')
    else:
        empty_res_element(net, element, suffix)


def get_relevant_elements(mode="pf"):
    if mode == "pf" or mode == "opf":
        return ["bus", "bus_dc", "line", "line_dc", "trafo", "trafo3w", "impedance", "ext_grid",
                "load", "motor", "sgen", "storage", "shunt", "gen", "ward",
                "xward", "dcline", "asymmetric_load", "asymmetric_sgen",
                "switch", "tcsc", "svc", "ssc", "vsc"]
    elif mode == "sc":
        return ["bus", "line", "trafo", "trafo3w", "ext_grid", "gen", "sgen", "switch"]
    elif mode == "se":
        return ["bus", "line", "trafo", "trafo3w", "impedance", "switch", "shunt"]
    elif mode == "pf_3ph":
        return ["bus", "line", "trafo", "ext_grid", "shunt",
                "load", "sgen", "storage", "asymmetric_load", "asymmetric_sgen"]


def init_results(net, mode="pf"):
    elements = get_relevant_elements(mode)
    suffix = suffix_mode.get(mode, None)
    for element in elements:
        init_element(net, element, suffix)


def reset_results(net, mode="pf"):
    elements = get_relevant_elements(mode)
    suffix = suffix_mode.get(mode, None)
    for element in elements:
        empty_res_element(net, element, suffix)
    if "res_cost" in net.keys():
        del net["res_cost"]


def _ppci_bus_to_ppc(result, ppc):
    # result is the ppci (ppc without out-of-service buses)
    # buses are sorted (REF, PV, PQ, NONE) -> results are the first 3 types
    for bus_table in ("bus", "bus_dc"):
        n_buses, bus_cols = np.shape(ppc[bus_table])
        n_rows_result, bus_cols_result = np.shape(result[bus_table])
        # create matrix of proper size
        updated_bus = np.empty((n_buses, bus_cols_result))
        # fill in results (first 3 types)
        updated_bus[:n_rows_result, :] = result[bus_table]
        if n_buses > n_rows_result:
            # keep rows for busses of type NONE
            updated_bus[n_rows_result:, :bus_cols] = ppc[bus_table][n_rows_result:, :]
        ppc[bus_table] = updated_bus

    ppc['svc'][result["internal"]['svc_is'], :] = result['svc'][:, :]
    ppc['ssc'][result["internal"]['ssc_is'], :] = result['ssc'][:, :]
    ppc['vsc'][result["internal"]['vsc_is'], :] = result['vsc'][:, :]


def _ppci_branch_to_ppc(result, ppc):
    # in service branches and gens are taken from 'internal'
    branch_cols = np.shape(ppc['branch'])[1]
    ppc['branch'][result["internal"]['branch_is'], :branch_cols] = result['branch'][:, :branch_cols]

    ppc['tcsc'][result["internal"]['tcsc_is'], :] = result['tcsc'][:, :]

    ppc['branch_dc'][result["internal"]['branch_dc_is'], :] = result['branch_dc'][:, :]


def _ppci_gen_to_ppc(result, ppc):
    gen_cols = np.shape(ppc['gen'])[1]
    ppc['gen'][result["internal"]['gen_is'], :gen_cols] = result['gen'][:, :gen_cols]


def _ppci_other_to_ppc(result, ppc, mode):

    if mode != "sc" and mode != "se":
        ppc['success'] = result['success']
        ppc['et'] = result['et']

    if mode == 'opf':
        ppc['obj'] = result['f']
        ppc['internal_gencost'] = result['gencost']

    if "iterations" in result:
        ppc["iterations"] = result["iterations"]


def _ppci_internal_to_ppc(result, ppc):
    for key, value in result["internal"].items():
        # Only for sc calculation
        # if branch current matrices have been stored they need to include out of service elements
        if key in BRANCH_RESULTS_KEYS:

            # n_buses = np.shape(ppc['bus'])[0]
            n_branches = np.shape(ppc['branch'])[0]
            # n_rows_result = np.shape(result['bus'])[0]
            # update_matrix = np.empty((n_branches, n_buses)) * np.nan
            # update_matrix[result["internal"]['branch_is'], :n_rows_result] = result["internal"][key]

            # To select only required buses and pad one column of nan value for oos bus
            update_matrix = np.empty((n_branches, value.shape[1]+1)) * 0.0
            update_matrix[result["internal"]['branch_is'],
                          :value.shape[1]] = result["internal"][key]
            ppc['internal'][key] = update_matrix
            if "br_res_ks_ppci_bus" in result["internal"]:
                br_res_ks_ppci_bus = np.r_[result["internal"]["br_res_ks_ppci_bus"], [-1]]
            else:
                br_res_ks_ppci_bus = np.r_[np.arange(value.shape[1]), [-1]]
            ppc['internal'][key] = pd.DataFrame(data=update_matrix, columns=br_res_ks_ppci_bus)
        else:
            ppc["internal"][key] = value


def _copy_results_ppci_to_ppc(result, ppc, mode):
    """
    result contains results for all in service elements
    ppc gets the results for in- and out of service elements
    -> results must be copied

    ppc and ppci are structured as follows:
          [in_service elements]
    ppc = [out_of_service elements]
    result = [in_service elements]

    Parameters
    ----------
    result - ppci with results
    ppc - ppc without results
    mode - "pf","opf", "sc"...

    Returns
    -------
    ppc with results
    """

    # copy the results for bus, gen and branch and some additional values like "success"
    _ppci_bus_to_ppc(result, ppc)
    _ppci_branch_to_ppc(result, ppc)
    _ppci_gen_to_ppc(result, ppc)
    _ppci_internal_to_ppc(result, ppc)
    _ppci_other_to_ppc(result, ppc, mode)

    result = ppc
    return result



import logging
import time

import pandas as pd

from pandapower.converter.cim import cim_tools
from pandapower.converter.cim.cim2pp import build_pp_net
from pandapower.converter.cim.other_classes import Report, LogLevel, ReportCode

logger = logging.getLogger('cim.cim2pp.converter_classes.asynchronousMachinesCim16')

sc = cim_tools.get_pp_net_special_columns_dict()


class AsynchronousMachinesCim16:
    def __init__(self, cimConverter: build_pp_net.CimConverter):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cimConverter = cimConverter

    def convert_asynchronous_machines_cim16(self):
        time_start = time.time()
        self.logger.info("Start converting AsynchronousMachines.")
        eqssh_asynchronous_machines = self._prepare_asynchronous_machines_cim16()
        self.cimConverter.copy_to_pp('motor', eqssh_asynchronous_machines)
        self.logger.info("Created %s motors in %ss." %
                         (eqssh_asynchronous_machines.index.size, time.time() - time_start))
        self.cimConverter.report_container.add_log(Report(
            level=LogLevel.INFO, code=ReportCode.INFO_CONVERTING,
            message="Created %s motors from AsynchronousMachines in %ss." %
                    (eqssh_asynchronous_machines.index.size, time.time() - time_start)))

    def _prepare_asynchronous_machines_cim16(self) -> pd.DataFrame:
        eqssh_generating_units = self.cimConverter.merge_eq_ssh_profile('WindGeneratingUnit')
        # a column for the type of the static generator in pandapower
        eqssh_generating_units['type'] = 'WP'
        eqssh_generating_units = pd.concat([eqssh_generating_units,
                                            self.cimConverter.merge_eq_ssh_profile('GeneratingUnit')],
                                           sort=False)
        eqssh_generating_units['type'] = eqssh_generating_units['type'].fillna('GeneratingUnit')
        eqssh_generating_units = pd.concat([eqssh_generating_units,
                                            self.cimConverter.merge_eq_ssh_profile('HydroGeneratingUnit')],
                                           sort=False)
        eqssh_generating_units['type'] = eqssh_generating_units['type'].fillna('Hydro')
        eqssh_generating_units = pd.concat([eqssh_generating_units,
                                            self.cimConverter.merge_eq_ssh_profile('SolarGeneratingUnit')],
                                           sort=False)
        eqssh_generating_units['type'] = eqssh_generating_units['type'].fillna('PV')
        eqssh_generating_units = pd.concat([eqssh_generating_units,
                                            self.cimConverter.merge_eq_ssh_profile('ThermalGeneratingUnit')],
                                           sort=False)
        eqssh_generating_units['type'] = eqssh_generating_units['type'].fillna('Thermal')
        eqssh_generating_units = pd.concat([eqssh_generating_units,
                                            self.cimConverter.merge_eq_ssh_profile('NuclearGeneratingUnit')],
                                           sort=False)
        eqssh_generating_units['type'] = eqssh_generating_units['type'].fillna('Nuclear')
        eqssh_generating_units = eqssh_generating_units.rename(columns={'rdfId': 'GeneratingUnit'})
        if 'sc' in self.cimConverter.cim.keys():
            asynchronous_machines = self.cimConverter.merge_eq_other_profiles(['ssh', 'sc'],
                                                                              'AsynchronousMachine',
                                                                              add_cim_type_column=True)
        else:
            asynchronous_machines = self.cimConverter.merge_eq_ssh_profile('AsynchronousMachine',
                                                                           add_cim_type_column=True)
        # prevent conflict of merging two dataframes each containing column 'name'
        eqssh_generating_units = eqssh_generating_units.drop('name', axis=1)
        asynchronous_machines = pd.merge(asynchronous_machines, eqssh_generating_units,
                                         how='left', suffixes=('_x', '_y'), on='GeneratingUnit')
        asynchronous_machines = pd.merge(asynchronous_machines, self.cimConverter.bus_merge, how='left',
                                         on='rdfId')
        asynchronous_machines['p_mw'] = -asynchronous_machines['p']
        asynchronous_machines['q_mvar'] = -asynchronous_machines['q']
        asynchronous_machines['current_source'] = True
        asynchronous_machines['cos_phi_n'] = asynchronous_machines['ratedPowerFactor'][:]
        asynchronous_machines['sn_mva'] = \
            asynchronous_machines['ratedS'].fillna(asynchronous_machines['nominalP'])
        asynchronous_machines['generator_type'] = 'async'
        asynchronous_machines['loading_percent'] = \
            100 * asynchronous_machines['p_mw'] / asynchronous_machines['ratedMechanicalPower']
        if 'inService_x' in asynchronous_machines.columns:
            asynchronous_machines['connected'] = (asynchronous_machines['connected']
                                                  & asynchronous_machines['inService_x']
                                                  & asynchronous_machines['inService_y'])
        asynchronous_machines = asynchronous_machines.rename(columns={'rdfId_Terminal': sc['t'], 'rdfId': sc['o_id'],
                                                                      'connected': 'in_service', 'index_bus': 'bus',
                                                                      'rxLockedRotorRatio': 'rx', 'iaIrRatio': 'lrc_pu',
                                                                      'ratedPowerFactor': 'cos_phi', 'ratedU': 'vn_kv',
                                                                      'efficiency': 'efficiency_n_percent',
                                                                      'ratedMechanicalPower': 'pn_mech_mw'})
        asynchronous_machines['scaling'] = 1
        asynchronous_machines['efficiency_percent'] = 100
        return asynchronous_machines
